from VLMP.components.models import modelBase

import copy

class SIMULATION(modelBase):
    """
    Component name: SIMULATION
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 6/11/2023

    Load model from dictionary generated by pyUAMMD simulation

    :param inputSimulation: Input dictionary generated by pyUAMMD simulation
    :type inputSimulation: dict
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"inputSimulation"},
                         requiredParameters  = {"inputSimulation"},
                         definedSelections   = {"particleId","forceField"},
                         **params)

        ############################################################

        inputSimulation = params["inputSimulation"]
        self.logger.info(f"[SIMULATION] Created model from pyUAMMD simulation")

        ########################################################

        # TYPES

        types = self.getTypes()

        typesLabels = inputSimulation["global"]["types"]["labels"]
        for typ in inputSimulation["global"]["types"]["data"]:
            typInfo = {l:typ[i] for i,l in enumerate(typesLabels)}
            types.addType(**typInfo)

        #Generate positions, a line along the z axis
        state = copy.deepcopy(inputSimulation["state"])

        #Generate structure
        structure = copy.deepcopy(inputSimulation["topology"]["structure"])

        #Generate forceField
        forceField = copy.deepcopy(inputSimulation["topology"]["forceField"])

        ########################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        if "forceField" in params:
            sel += self.getForceFieldSelection(params["forceField"])

        return sel



import sys,os

import logging

import json

import argparse

from .utils.launcher import *

if __name__ == "__main__":

    #Create argument parser
    mainParser = argparse.ArgumentParser(description='Run a set of simulations')

    #Add arguments
    mainParser.add_argument('--session', type=str, help='VLMP session info', required=True)

    #There are two simualtions options, local and liquid (cluster lsf). If none is selected, local is used.
    #Add mutually exclusive group
    group = mainParser.add_mutually_exclusive_group()
    group.add_argument('--rebuild' , action='store_true', help='Rebuild results')
    group.add_argument('--local'   , action='store_true', help='Run simulations locally')
    group.add_argument('--liquid'  , action='store_true', help='Run simulations in liquid cluster')
    group.add_argument('--slurm'   , action='store_true', help='Run simulations in slurm cluster')

    mainArgs,_ = mainParser.parse_known_args()

    parser = argparse.ArgumentParser(parents=[mainParser],add_help=False)
    if mainArgs.local:
        parser.add_argument('--gpu', nargs='+', type=int, help='List of gpu ids to use',required=True)

    if mainArgs.liquid:
        parser.add_argument('--node', nargs='+', type=str, help='List of node ids to use',required=True)
        parser.add_argument('--postScript', type=str, help='Post script to run after simulation',required=False)

    if mainArgs.slurm:
        parser.add_argument('--node', nargs='+', type=str, help='List of node ids to use',required=False)
        parser.add_argument('--partition', nargs='+', type=str, help='List of node ids to use',required=True)
        parser.add_argument('--filling',nargs='+', type=int, help='List of sim per node',required=False)
        parser.add_argument('--modules', nargs='+', type=str, help='List of modules used',required=False)
        parser.add_argument('--postScript', type=str, help='Post script to run after simulation',required=False)

    #Parse arguments
    args = parser.parse_args()

    #########################################################

    #Load simulation sets info
    with open(args.session,"r") as f:
        simulationSetsInfo = json.load(f)

    #Start VLMP

    logger = logging.getLogger("VLMP")

    #Add file handler to logger
    fileHandler = logging.FileHandler(simulationSetsInfo["name"]+".log","w")
    #Set the same level as the logger
    fileHandler.setLevel(logger.level)
    #Get the same formatter as the logger
    fileHandler.setFormatter(logger.handlers[0].formatter)

    #Add file handler to logger
    logger.addHandler(fileHandler)

    logger.info("Starting VLMP ...")

    if mainArgs.rebuild:
        logger.info("Rebuilding results folders ...")
        rebuildResults(simulationSetsInfo)

    elif mainArgs.local:
        #Remove console handler
        logger.removeHandler(logger.handlers[0])

        logger.info("Running simulations locally ...")
        child_pid = os.fork()

        if child_pid == 0:
            localLauncher(simulationSetsInfo,args.gpu)
        else:
            sys.exit(0)
    elif mainArgs.liquid:
        #Remove console handler
        logger.removeHandler(logger.handlers[0])

        logger.info("Running simulations in liquid cluster ...")

        postScript = args.postScript if args.postScript else ""
        liquidLauncher(simulationSetsInfo,args.node,postScript)
    elif mainArgs.slurm:
        #Remove console handler
        logger.removeHandler(logger.handlers[0])

        logger.info("Running simulations in slurm cluster ...")

        if args.node:
            node       = args.node
            filling    = args.filling if args.filling else [1]*len(args.node)
        else:
            if args.filling:
                logger.error("Filling only can be used when a node list is given")
                sys.exit(1)
            node    = []
            filling = []

        if len(node) != len(filling):
            logger.error("The size of node and filling must match")
            sys.exit(1)

        partition  = args.partition
        modules    = args.modules if args.modules else [""]
        postScript = args.postScript if args.postScript else ""
        slurmLauncher(simulationSetsInfo,node,filling,partition,modules,postScript)

    else:
        logger.error("No simulation option selected")
        sys.exit(1)

    sys.exit(0)



from VLMP.components.models import modelBase

import numpy as np
#from scipy.spatial import cKDTree

from icosphere import icosphere

class SPHEREMULTIBLOB(modelBase):
    """
    Component name: SPHEREMULTIBLOB
    Component type: model

    Author: Pablo Ibáñez-Freire and Pablo Palacios-Alonso
    Date: 18/07/2023

    Extension of Icosidodecahedron + icosphere
    """

    def __even_permutation(self, a, b, c, pos, i1):
        pos[i1, :] = [a, b, c]
        pos[i1 + 1, :] = [c, a, b]
        pos[i1 + 2, :] = [b, c, a]

    def __vertex_distances(self, NV, pos, iout):
        dmin = 10. ** 10
        for i in range(NV):
            for j in range(i + 1, NV):
                r1 = pos[i, :]
                r2 = pos[j, :]
                dist = self.__distance(r1, r2)
                if dist < dmin:
                    dmin = dist
        return dmin

    def __distance(self, r1, r2):
        dist = np.sum((r1 - r2) ** 2)
        return np.sqrt(dist)

    def __radius(self, nv, pos):
        r2 = pos[nv, :]  # center
        dists = np.zeros(nv, dtype=float)
        for i in range(nv):
            r1 = pos[i, :]
            dists[i] = self.__distance(r1, r2)
        return np.max(dists)


    def __sphere(self, rSphere):

        if self.sphereType == "icosphere":

            mu = int(np.sqrt(0.1*(self.sphN-12)+1))
            N    = 12+10*(mu**2-1)
            Nmax = 12+10*((mu+1)**2-1)

            if self.sphN != N:
                self.logger.error(f"The given number of spheres for sphere multiblob ({self.sphN}) is not valid for icosphere."
                                  f"The lower valid number is {N} and the upper valid number is {Nmax}.")
                raise Exception("Invalid number of spheres for icosphere")

            pos, _ = icosphere(mu)
            pos    = np.append(pos, [[0, 0, 0]], axis=0)
            pos   *= rSphere

            edgeIco    = rSphere/np.sin(2*np.pi/5)
            edgeLength = edgeIco/mu

        elif self.sphereType == "icosidodecahedron":

            if self.sphN != 30:
                self.logger.error("Number of particles for icosidodecahedron must be 30")
                raise Exception("Number of particles for icosidodecahedron must be 30")

            goldenRatio = (1.0 + np.sqrt(5.0))/2.0

            pos = np.zeros((self.sphN+1, 3), dtype=float)
            ic  = np.zeros(self.sphN+1, dtype=int)
            edgeLength = rSphere / goldenRatio

            a = 0
            b = 0
            c = goldenRatio
            cm = -c
            self.__even_permutation(a, b, c, pos, 0)
            self.__even_permutation(a, b, cm, pos, 3)

            a = 0.5
            b = goldenRatio / 2.0
            c = (goldenRatio + 1.) / 2.
            am = -a
            bm = -b
            cm = -c

            self.__even_permutation(a, b, c, pos, 6)
            self.__even_permutation(am, b, c, pos, 9)
            self.__even_permutation(a, bm, c, pos, 12)
            self.__even_permutation(a, b, cm, pos, 15)
            self.__even_permutation(am, bm, c, pos, 18)
            self.__even_permutation(am, b, cm, pos, 21)
            self.__even_permutation(a, bm, cm, pos, 24)
            self.__even_permutation(am, bm, cm, pos, 27)

            pos[30, :] = 0  # central
            ic[0] = 2
            ic[6] = 3
            ic[9] = 3
            ic[12] = 3
            ic[18] = 3
            pos = edgeLength * pos

        else:
            self.logger.error("Sphere type not recognized")
            raise Exception("Sphere type not recognized")

        return pos, edgeLength

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"sphereType",
                                                "particleName",
                                                "particleMass","particleRadius","particleCharge",
                                                "numberOfSpheres",
                                                "particlesPerSphere",
                                                "radiusOfSphere",
                                                "K",
                                                "heightMean","heightStd",
                                                "heightReference"},
                         requiredParameters  = {"K"},
                         definedSelections   = {"particleId"},
                         **params)

        ############################################################

        self.sphereType  = params.get("sphereType", "icosidodecahedron")

        if self.sphereType not in ["icosphere","icosidodecahedron"]:
            self.logger.error("Sphere type not recognized")
            raise Exception("Sphere type not recognized")

        self.sphN        = params.get("particlesPerSphere",31)

        particleName = params.get("particleName","A")

        particleMass   = params.get("particleMass",1.0)
        particleRadius = params.get("particleRadius",1.0)
        particleCharge = params.get("particleCharge",0.0)

        types = self.getTypes()
        types.addType(name = particleName,
                      mass = particleMass,
                      radius = particleRadius,
                      charge = particleCharge)

        numberOfSpheres = params.get("numberOfSpheres",1)
        radiusOfSphere  = params.get("radiusOfSphere",1.0)

        K = params["K"]

        heightMean = params.get("heightMean",0.0)
        heightStd  = params.get("heightStd",0.0)

        heightReference = params.get("heightReference",0.0)

        box = self.getEnsemble().getEnsembleComponent("box")

        self.maxTries    = params.get("maxTries",100)

        #TODO: apply pbc, now box size is reduced by 2*radiusOfSphere

        X = (box[0]-2*radiusOfSphere)/2
        Y = (box[1]-2*radiusOfSphere)/2

        #X = box[0]/2.0
        #Y = box[1]/2.0

        #Check box height and genration height
        Z = box[2]/2.0
        if heightReference > Z or heightReference < -Z:
            self.logger.error(f"Height reference is out of box {heightReference} > {Z} or {heightReference} < {-Z}")
            raise ValueError("Height reference is out of box")

        if heightMean + heightReference > Z or heightMean + heightReference < -Z:
            self.logger.error(f"Generation height is out of box {heightMean + heightReference} > {Z} or "
                              f"{heightMean + heightReference} < {-Z}")
            raise ValueError("Generation height is out of box")

        ############################################################

        sphPositions = []

        i=0
        tries=0
        while i < numberOfSpheres:
            tries+=1
            if tries >= self.maxTries*numberOfSpheres:
                self.logger.error("Unable to find a correct configuration")
                raise ValueError("The number of spheres is too high for the box size")
            if heightStd > 0.0:
                height = np.random.normal(heightMean,heightStd)
            else:
                height = heightMean
            height += heightReference

            x = np.random.uniform(-X,X)
            y = np.random.uniform(-Y,Y)

            center = [x,y,height]

            if height > Z - radiusOfSphere or height < -Z + radiusOfSphere:
                continue

            if sphPositions:
                #minDst,minDstIndex = cKDTree(sphPositions).query(center, 1)
                minDst = np.inf
                for p in sphPositions:
                    dst = np.linalg.norm(p-np.asarray(center))
                    minDst = min(dst,minDst)

            else:
                minDst = np.inf

            if minDst > 2.0*radiusOfSphere*1.1:
                sphPositions.append(center)
                i+=1

        # Generate spheres

        state = {}
        state["labels"]=["id","position"]
        state["data"]  =[]

        idCounter = 0
        sph2ids = []
        sph2pos = []
        for center in sphPositions:
            pos, edgeLength = self.__sphere(radiusOfSphere)
            pos += center
            ids = []
            for p in pos:
                state["data"].append([idCounter,list(p)])
                ids.append(idCounter)
                idCounter += 1
            sph2ids.append(ids.copy())
            sph2pos.append(pos.copy())

        structure = {}
        structure["labels"] = ["id", "type", "modelId"]
        structure["data"]   = []

        for i in range(numberOfSpheres):
            for j in sph2ids[i]:
                structure["data"].append([j,particleName,i])

        forceField = {}
        forceField["Bond"] = {}
        forceField["Bond"]["parameters"] = {}
        forceField["Bond"]["type"] = ["Bond2", "Harmonic"]
        forceField["Bond"]["labels"] = ["id_i", "id_j", "K", "r0"]
        forceField["Bond"]["data"] = []

        bonds = []
        for i in range(numberOfSpheres):
            for n in range(self.sphN):
                for m in range(n + 1, self.sphN):
                    pn = sph2pos[i][n]
                    pm = sph2pos[i][m]

                    dst = np.linalg.norm(pn - pm)

                    if dst < 1.25 * edgeLength:
                        bonds.append([sph2ids[i][n], sph2ids[i][m], K, edgeLength])

            #All particles are bonded to central particle, which is the last one
            for j in sph2ids[i][:-1]:
                bonds.append([sph2ids[i][-1], j, K, radiusOfSphere])

        for i,j,k,r0 in bonds:
            forceField["Bond"]["data"].append([i,j,k,r0])

        ############################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)


    def processSelection(self,**params):

        sel = []
        if "particleId" in params:
            sel += params["particleId"]
        return sel




import sys, os

import logging

import numpy as np

from . import modelExtensionBase

class sphericalShell(modelExtensionBase):

    """
    Component name: sphericalShell
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 15/06/2023

    Spherical shell model extension for the model.

    :param shellCenter: Center of the spherical shell.
    :type shellCenter: list of floats
    :param shellRadius: Radius of the spherical shell.
    :type shellRadius: float
    :param shellEpsilon: Epsilon of the spherical shell.
    :type shellEpsilon: float, optional (default = 1.0)
    :param shellSigma: Sigma of the spherical shell.
    :type shellSigma: float, optional (default = 1.0)
    :param minShellRadius: Minimum radius of the spherical shell.
    :type minShellRadius: float, optional (default = 0.0)
    :param maxShellRadius: Maximum radius of the spherical shell.
    :type maxShellRadius: float, optional (default = inf)
    :param radiusVelocity: Velocity of the radius of the spherical shell.
    :type radiusVelocity: float, optional (default = 0.0)

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"shellCenter",
                                                "shellRadius",
                                                "shellEpsilon",
                                                "shellSigma",
                                                "minShellRadius",
                                                "maxShellRadius",
                                                "radiusVelocity"},
                         requiredParameters  = {"shellCenter",
                                                "shellRadius"},
                         availableSelections = {"selection"},
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        shellCenter = params["shellCenter"]
        shellRadius = params["shellRadius"]

        ###########

        shellEpsilon = params.get("shellEpsilon",1.0)
        shellSigma   = params.get("shellSigma",1.0)

        minShellRadius = params.get("minShellRadius",0.0)
        maxShellRadius = params.get("maxShellRadius","INFINITY")

        radiusVelocity = params.get("radiusVelocity",0.0)

        ###########

        if shellRadius == "auto":
            if "selection" not in params:
                self.logger.error(f"[sphericalShell] ({name}) Selection \
                                    is required when shellRadius is 'auto'.")
                raise Exception("Not compatible parameters.")

            # Get selection
            selIds = self.getSelection("selection")

            # Get the maximum distance to the center
            posIds = self.getIdsState(selIds, "position")
            maxDist = np.max(np.linalg.norm(np.asarray(posIds) - shellCenter, axis=1))

            self.logger.debug(f"[sphericalShell] ({name}) Maximum distance to the center: {maxDist}")

            # Get the maximum radius
            radIds = self.getIdsProperty(selIds, "radius")
            maxRad = np.max(radIds)

            self.logger.debug(f"[sphericalShell] ({name}) Maximum radius: {maxRad}")

            shellRadius = maxDist + (shellSigma  + maxRad)*1.1

            self.logger.info(f"[sphericalShell] ({name}) Shell radius: {shellRadius}")

        elif "selection" in params:
            self.logger.error(f"[sphericalShell] ({name}) Selection selected \
                                but radius is not set to 'auto'.")
            raise Exception("Not compatible parameters.")

        ###########

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["External","SphericalShell"]
        extension[name]["parameters"] = {"shellCenter":shellCenter,
                                         "shellRadius":shellRadius,
                                         "shellEpsilon":shellEpsilon,
                                         "shellSigma":shellSigma,
                                         "minShellRadius":minShellRadius,
                                         "radiusVelocity":radiusVelocity}

        if maxShellRadius != "INFINITY":
            extension[name]["parameters"]["maxShellRadius"] = maxShellRadius

        ############################################################

        self.setExtension(extension)



import os
import logging

import copy

from tqdm import tqdm

import json
import jsbeautifier

from collections import OrderedDict

from . import DEBUG_MODE

import VLMP.components.systems         as _systems
import VLMP.components.units           as _units
import VLMP.components.types           as _types
import VLMP.components.ensembles       as _ensembles
import VLMP.components.models          as _models
import VLMP.components.modelOperations as _modelOperations
import VLMP.components.modelExtensions as _modelExtensions
import VLMP.components.integrators     as _integrators
import VLMP.components.simulationSteps as _simulationSteps

from pyUAMMD.utils.merging.merging import mergeSimulationsSet

import importlib
import inspect

class VLMP:

    def __setUpAdditionalComponents(self,additionalComponets = None):

        for component in self.availableComponents:
            compName = "additional" + component[0].upper() + component[1:]
            setattr(self,compName,None)

        if additionalComponets is not None:

            if os.path.isdir(additionalComponets):
                self.logger.info("[VLMP] Loading additional components from \"%s\"",additionalComponets)
            else:
                #Create additional components folder
                self.logger.info("[VLMP] Creating additional components folder \"%s\"",additionalComponets)
                os.mkdir(additionalComponets)

            #Check a folder for each available component exists, else create it
            for component in self.availableComponents:
                componentPath = os.path.join(additionalComponets,component)
                if not os.path.isdir(componentPath):
                    self.logger.info("[VLMP] Creating additional components folder \"%s\"",componentPath)
                    os.mkdir(componentPath)

            #Check a file named "__init__.py" exists in each folder, else create it
            for component in self.availableComponents:
                componentPath = os.path.join(additionalComponets,component,"__init__.py")
                if not os.path.isfile(componentPath):
                    self.logger.info("[VLMP] Creating __init__.py for additional components folder \"%s\"",componentPath)
                    open(componentPath,"w").close()

            for component in self.availableComponents:
                compName = "additional" + component[0].upper() + component[1:]
                setattr(self,compName,importlib.import_module(".".join([additionalComponets,component])))

            #List all additional components imported
            for component in self.availableComponents:
                compName = "additional" + component[0].upper() + component[1:]
                comp = [name for name, obj in vars(eval(f"self.{compName}")).items() if not name.startswith('_') and name != f"{component}Base" and inspect.isclass(obj)]
                self.logger.info("[VLMP] Additional components \"%s\": %s",component,comp)

        self.componentsModules["system"]         ["additional"] = self.additionalSystem
        self.componentsModules["units"]          ["additional"] = self.additionalUnits
        self.componentsModules["types"]          ["additional"] = self.additionalTypes
        self.componentsModules["ensemble"]       ["additional"] = self.additionalEnsemble
        self.componentsModules["models"]         ["additional"] = self.additionalModels
        self.componentsModules["modelOperations"]["additional"] = self.additionalModelOperations
        self.componentsModules["modelExtensions"]["additional"] = self.additionalModelExtensions
        self.componentsModules["integrators"]    ["additional"] = self.additionalIntegrators
        self.componentsModules["simulationSteps"]["additional"] = self.additionalSimulationSteps

    ########################################

    #Distribute functions

    def __distributeSimulationPoolByMaxNumberOfParticles(self,maxNumberOfParticles):

        simulationSets = []

        for simSet in self.simulationSets:

            currentSet     = []
            currentSetSize = 0

            for simName in simSet:
                sim = self.simulations[simName]
                if currentSetSize + sim.getNumberOfParticles() > maxNumberOfParticles:
                    if len(currentSet) > 0:
                        simulationSets.append(currentSet)
                    currentSet     = []
                    currentSetSize = 0

                currentSet.append(simName)
                currentSetSize += sim.getNumberOfParticles()

            if len(currentSet) > 0:
                simulationSets.append(currentSet)

        #Print the number of simulations and the number of particles in each set
        for i in range(len(simulationSets)):
            self.logger.debug("[VLMP] Simulation set %d has %d simulations and %d particles (max %d)",
                              i,len(simulationSets[i]),
                              sum([self.simulations[simName].getNumberOfParticles() for simName in simulationSets[i]]),
                              maxNumberOfParticles)

        return simulationSets.copy()

    def __distributeSimulationPoolBySize(self,size):
        simulationSets = []

        for simSet in self.simulationSets:

            currentSet     = []
            currentSetSize = 0

            for simName in simSet:
                sim = self.simulations[simName]
                if currentSetSize + 1 > size:
                    if len(currentSet) > 0:
                        simulationSets.append(currentSet)
                    currentSet     = []
                    currentSetSize = 0

                currentSet.append(simName)
                currentSetSize += 1

            if len(currentSet) > 0:
                simulationSets.append(currentSet)

        #Print the number of simulations and the number of particles in each set
        for i in range(len(simulationSets)):
            self.logger.debug("[VLMP] Simulation set %d has %d simulations",
                              i,len(simulationSets[i]))

        return simulationSets.copy()

    def __distributeSimulationPoolByProperty(self,propertyPath):

        simulationSets = []

        for simSet in self.simulationSets:

            simulationSetsProp = {}

            for simName in simSet:
                sim = self.simulations[simName]
                #Get the property value
                try:
                    propertyValue = sim[propertyPath[0]]
                    for i in range(1,len(propertyPath)):
                        propertyValue = propertyValue[propertyPath[i]]
                except:
                    self.logger.error("[VLMP] Property \"%s\" not found in simulation",propertyPath)
                    raise Exception("Property not found in simulation")

                if propertyValue not in simulationSetsProp.keys():
                    simulationSetsProp[propertyValue] = []

                simulationSetsProp[propertyValue].append(simName)

            simulationSets += simulationSetsProp.values()

        #Print the number of simulations and the number of particles in each set
        for i in range(len(simulationSets)):
            self.logger.debug("[VLMP] Simulation set %d has %d simulations",
                              i,len(simulationSets[i]))

        return simulationSets.copy()

    ########################################

    def __checkComponent(self,component,componentClass,simulationBuffer):
        componentType = component.get("type",None)

        if componentType is None:
            self.logger.error(f"[VLMP] Error processing \"{componentClass}\" component ({component}), \"type\" property not found")
            raise Exception("Component type not specified")
        self.logger.debug(f"[VLMP] Processing \"{componentClass}\" component ({component}), type \"{componentType}\"")

        componentName = component.get("name",None)
        if componentName is None:
            componentName = componentType
            self.logger.warning(f"[VLMP] ({componentClass}) Component name not specified, using \"{componentName}\".")
        self.logger.debug(f"[VLMP] ({componentClass}) Component name \"{componentName}\".")

        #Check if parameters are specified
        componentParameters = component.get("parameters",None)
        if componentParameters is None:
            componentParameters = {}
            self.logger.warning(f"[VLMP] ({componentClass}) Component parameters not specified, creating empty dictionary")
        self.logger.debug(f"[VLMP] ({componentClass}) Component parameters \"{componentParameters}\"")

        #Check if component is already loaded
        if "_".join([componentClass,componentName]) in simulationBuffer.keys():
            self.logger.error(f"[VLMP] ({componentClass}) Component \"{componentName}\" already loaded")
            raise Exception("Component already loaded")

        return componentType,componentName,componentParameters

    def __processSimulationPoolSection(self,
                                       simulationBuffer,
                                       simulationInfo,
                                       sectionName,
                                       addToSimulationBuffer,
                                       required,
                                       unique,
                                       units,types,ensemble,models):

        sectionNameUpper  = sectionName[0].upper()+sectionName[1:]

        if sectionName[-1] != "s":
            sectionNamePlural = sectionName+"s"
        else:
            sectionNamePlural = sectionName

        loadedComponents = []

        #Check if "sectionName" section is present
        if sectionName not in simulationInfo.keys():
            if not required:
                self.logger.warning(f"[VLMP] ({sectionName}) Section not found, skipping")
                return loadedComponents
            self.logger.error(f"[VLMP] {sectionNameUpper} section not found")
            raise Exception(f"{sectionNameUpper} section not found")
        else:

            if unique:
                #Only one unit system can be specified
                if len(simulationInfo[sectionName]) > 1:
                    self.logger.error(f"[VLMP] Only one {sectionName} entry can be specified")
                    raise Exception("Only one entry can be specified")

            for comp in simulationInfo[sectionName]:

                typ, name, param = self.__checkComponent(comp,sectionName,simulationBuffer)
                self.logger.debug(f"[VLMP] Adding {sectionName} \"{name}\"")

                isComp           = typ in dir(self.componentsModules[sectionName]["base"])
                isAdditionalComp = typ in dir(self.componentsModules[sectionName]["additional"])

                if not isComp and not isAdditionalComp:
                    self.logger.error(f"[VLMP] {sectionNameUpper} \"{typ}\" not found")
                    raise Exception(f"{sectionNameUpper} not found")

                if isComp and isAdditionalComp:
                    self.logger.error(f"[VLMP] {sectionNameUpper} \"{typ}\" found in both VLMP base and additional components")
                    raise Exception(f"{sectionNameUpper} found in both VLMP base and additional components")

                allArgs = {
                    'name'    : name,
                    'units'   : units,
                    'types'   : types,
                    'ensemble': ensemble,
                    'models'  : models
                }

                args = {k: v for k, v in allArgs.items() if v is not None}

                if isComp:

                    try:

                        initComp = eval(f"_{sectionNamePlural}.{typ}")(**args, **param)

                        if addToSimulationBuffer:
                            simulationBuffer[f"{sectionNamePlural}_{name}"] = initComp

                        loadedComponents.append(f"{sectionNamePlural}_{name}")
                    except:
                        self.logger.error(f"[VLMP] Error loading {sectionName} \"{name}\" ({typ})")
                        raise Exception(f"Error loading {sectionName}")

                if isAdditionalComp:
                    try:

                        initComp = eval(f"self.additional{sectionNameUpper}.{typ}")(**args,**param)

                        if addToSimulationBuffer:
                            simulationBuffer[f"{sectionName}_{name}"] = initComp

                        loadedComponents.append(f"{sectionName}_{name}")
                    except:
                        self.logger.error(f"[VLMP] Error loading {sectionName} \"{name}\" ({typ})")
                        raise Exception(f"Error loading {sectionName}")

        return loadedComponents

    ########################################

    def __init__(self,additionalComponets = None):
        self.logger = logging.getLogger("VLMP")

        self.logger.info("[VLMP] Starting VLMP")

        self.simulations    = OrderedDict()
        self.simulationSets = []

        self.availableComponents = ["system","units","types","ensemble",
                                    "models","modelOperations","modelExtensions",
                                    "integrators",
                                    "simulationSteps"]

        self.componentsModules = {"system"          : {"base":_systems        },
                                  "units"           : {"base":_units          },
                                  "types"           : {"base":_types          },
                                  "ensemble"        : {"base":_ensembles      },
                                  "models"          : {"base":_models         },
                                  "modelOperations" : {"base":_modelOperations},
                                  "modelExtensions" : {"base":_modelExtensions},
                                  "integrators"     : {"base":_integrators    },
                                  "simulationSteps" : {"base":_simulationSteps}}

        self.__setUpAdditionalComponents(additionalComponets)

    def loadSimulationPool(self,simulationPool:list):

        self.simulationsInfo = {}

        for simulationInfo in simulationPool:

            #Check all keys are available components
            for key in simulationInfo.keys():
                if key not in self.availableComponents:
                    self.logger.error("[VLMP] Unknown component \"%s\"",key)
                    self.logger.error("[VLMP] Available components are: %s",self.availableComponents)
                    raise Exception("Unknown component")

            simulationBuffer = OrderedDict()

            ############## SYSTEM ##############

            #### Get simulation name ####

            # Check there is one (and only one) system component of type "simulationName"
            simNameComponents = [component for component in simulationInfo["system"] if component["type"] == "simulationName"]
            if len(simNameComponents) == 0:
                self.logger.error("[VLMP] Simulation name not specified")
                raise Exception("Simulation name not specified")
            elif len(simNameComponents) > 1:
                self.logger.error("[VLMP] More than one simulation name specified")
                raise Exception("More than one simulation name specified")

            for comp in simulationInfo["system"]:

                typ, name, param = self.__checkComponent(comp,"system",simulationBuffer)

                #Read simulationName
                if typ == "simulationName":
                    simulationName = param["simulationName"]

                    #Check if other simulation with the same name has been already created
                    if simulationName in self.simulations.keys():
                        self.logger.error(f"[VLMP] Simulation with name \"{simulationName}\" already exists")
                        raise Exception("Simulation already exists")

            #############################

            _ = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                    simulationInfo   = simulationInfo,
                                                    sectionName      = "system",
                                                    addToSimulationBuffer = True,
                                                    required = True,
                                                    unique   = False,
                                                    units    = None,
                                                    types    = None,
                                                    ensemble = None,
                                                    models   = None)

            ############## UNITS ##############

            units = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                        simulationInfo = simulationInfo,
                                                        sectionName    = "units",
                                                        addToSimulationBuffer = True,
                                                        required = True,
                                                        unique   = False,
                                                        units    = None,
                                                        types    = None,
                                                        ensemble = None,
                                                        models   = None)

            units = simulationBuffer[units[0]]

            ############## TYPES ##############

            types = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                        simulationInfo = simulationInfo,
                                                        sectionName    = "types",
                                                        addToSimulationBuffer = True,
                                                        required = True,
                                                        unique   = False,
                                                        units    = units,
                                                        types    = None,
                                                        ensemble = None,
                                                        models   = None)

            types = simulationBuffer[types[0]]

            ############## ENSEMBLE ##############

            ensemble = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                           simulationInfo = simulationInfo,
                                                           sectionName    = "ensemble",
                                                           addToSimulationBuffer = True,
                                                           required = True,
                                                           unique   = False,
                                                           units    = units,
                                                           types    = types,
                                                           ensemble = None,
                                                           models   = None)

            ensemble = simulationBuffer[ensemble[0]]

            ############### MODEL ###############

            models = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                         simulationInfo = simulationInfo,
                                                         sectionName    = "models",
                                                         addToSimulationBuffer = True,
                                                         required = True,
                                                         unique   = False,
                                                         units    = units,
                                                         types    = types,
                                                         ensemble = ensemble,
                                                         models   = None)

            #Set idOffset for each model
            idOffset = 0
            for mdl in models:
                simulationBuffer[mdl].setIdOffset(idOffset)
                ids = simulationBuffer[mdl].getLocalIds()
                if len(ids) != 0:
                    idOffset += max(ids) + 1

            models = [simulationBuffer[model] for model in models]

            ############### MODEL OPERATIONS ###############

            _ = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                    simulationInfo = simulationInfo,
                                                    sectionName    = "modelOperations",
                                                    addToSimulationBuffer = False,
                                                    required = False,
                                                    unique   = False,
                                                    units    = units,
                                                    types    = types,
                                                    ensemble = ensemble,
                                                    models   = models)

            ############### MODEL EXTENSIONS ###############

            _ = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                    simulationInfo = simulationInfo,
                                                    sectionName    = "modelExtensions",
                                                    addToSimulationBuffer = True,
                                                    required = False,
                                                    unique   = False,
                                                    units    = units,
                                                    types    = types,
                                                    ensemble = ensemble,
                                                    models   = models)

            ############## INTEGRATOR ##############

            _ = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                    simulationInfo = simulationInfo,
                                                    sectionName    = "integrators",
                                                    addToSimulationBuffer = True,
                                                    required = True,
                                                    unique   = False,
                                                    units    = units,
                                                    types    = types,
                                                    ensemble = ensemble,
                                                    models   = models)

            ############### SIMULATION STEPS ###############

            _ = self.__processSimulationPoolSection(simulationBuffer = simulationBuffer,
                                                    simulationInfo = simulationInfo,
                                                    sectionName    = "simulationSteps",
                                                    addToSimulationBuffer = True,
                                                    required = False,
                                                    unique   = False,
                                                    units    = units,
                                                    types    = types,
                                                    ensemble = ensemble,
                                                    models   = models)

            ###############################################

            #Merge all components into a single simulation
            self.logger.debug("[VLMP] Merging components into a single simulation")

            sim = None
            for componentName,component in simulationBuffer.items():
                self.logger.debug(f"[VLMP] Merging component \"{componentName}\"")
                if sim is None:
                    sim = component.getSimulation(DEBUG_MODE)
                else:
                    sim.append(component.getSimulation(DEBUG_MODE),mode="modelId")
                self.logger.debug(f"[VLMP] Component \"{componentName}\" merged")
            #Simulation creation finished

            ###############################################

            #Store the simulation
            self.simulationsInfo[simulationName] = copy.deepcopy(simulationInfo)
            self.simulations[simulationName]     = sim

        ###############################################

        #At this point simulation pool is processed

        #Create default simulation set, all simulations in one set
        self.simulationSets = [list(self.simulations.keys())]

    def distributeSimulationPool(self,*mode):

        availableModes = ["none","one","upperLimit","size","property"]

        #Check at least one simulations has been loaded
        if len(self.simulations) == 0:
            self.logger.error("[VLMP] No simulations loaded")
            raise Exception("No simulations loaded")

        #Check mode
        if len(mode) == 0:
            self.logger.warning("[VLMP] No mode specified, using \"none\"")
            modeName = "none"
        else:
            modeName = mode[0]
            self.logger.debug(f"[VLMP] Distributing simulation pool using mode \"{modeName}\"")

        if modeName not in availableModes:
            self.logger.error("[VLMP] Distribute mode \"%s\" not available, available modes are: %s",modeName,availableModes)
            raise Exception("Distribute mode not available")
        else:
            #Switch to the selected mode
            if modeName == "none":
                pass
            elif  modeName == "one":
                self.simulationSets = [[i] for i in self.simulations.keys()]
            elif modeName == "upperLimit":
                self.logger.debug("[VLMP] Distributing simulation pool using upper limit")
                availableScoringProperties = ["numberOfParticles"]

                if len(mode) >= 2:
                    scoringPropertyName = mode[1]
                else:
                    self.logger.error("[VLMP] No scoring property specified")
                    raise Exception("No scoring property specified")

                if scoringPropertyName not in availableScoringProperties:
                    self.logger.error("[VLMP] Scoring property \"%s\" not available, available properties are: %s",
                                      scoringPropertyName,availableScoringProperties)
                    raise Exception("Scoring property not available")
                else:
                    #Switch to the selected scoring property
                    if scoringPropertyName == "numberOfParticles":
                        self.logger.debug("[VLMP] Distributing simulation pool using number of particles")
                        if len(mode) >= 3:
                            maxNumberOfParticles = mode[2]
                        else:
                            self.logger.error("[VLMP] No particle limit specified")
                            raise Exception("No upper limit specified")

                        #Distribute the simulation pool
                        self.simulationSets = self.__distributeSimulationPoolByMaxNumberOfParticles(maxNumberOfParticles)
                    #Scoring property switch finished
            elif modeName == "size":
                self.logger.debug("[VLMP] Distributing simulation pool using size")

                if len(mode) >= 2:
                    size = mode[1]
                else:
                    self.logger.error("[VLMP] No size specified")
                    raise Exception("No size specified")

                #Distribute the simulation pool
                self.simulationSets = self.__distributeSimulationPoolBySize(size)

            elif modeName == "property":
                self.logger.debug("[VLMP] Distributing simulation pool using property")

                if len(mode) >= 2:
                    propertyPath = mode[1]
                    #Check if property path is valid is list of strings
                    if not isinstance(propertyPath,list):
                        self.logger.error("[VLMP] Property path must be a list of strings")
                        raise Exception("Property path must be a list of strings")
                    else:
                        for property in propertyPath:
                            if not isinstance(property,str):
                                self.logger.error("[VLMP] Property path must be a list of strings")
                                raise Exception("Property path must be a list of strings")
                else:
                    self.logger.error("[VLMP] No property path specified")
                    raise Exception("No scoring property specified")

                #Distribute the simulation pool
                self.simulationSets = self.__distributeSimulationPoolByProperty(propertyPath)


        #Check all the simulations have been distributed.
        #simulationSets is a list of lists which contains the names of the simulations
        #in the simulation pool

        simulationNames       = list(self.simulations.keys())
        simulationNamesInSets = [s for st in self.simulationSets for s in st]

        #Both must be equal
        if sorted(simulationNames) != sorted(simulationNamesInSets):
            self.logger.error("[VLMP] Simulation distribution failed")
            raise Exception("Simulation distribution failed")

    def setUpSimulation(self, sessionName):
        self.logger.debug("[VLMP] Setting up simulation")

        if len(self.simulationSets) == 0:
            self.logger.error("[VLMP] Simulation pool not distributed")
            raise Exception("Simulation pool not distributed")

        ################################################

        #Create folder named sessionName
        if not os.path.exists(sessionName):
            os.makedirs(sessionName)

        ################################################

        #Create folder sessionName/simulationSets
        if not os.path.exists(os.path.join(sessionName,"simulationSets")):
            os.makedirs(os.path.join(sessionName,"simulationSets"))

        #Create folder sessionName/results
        if not os.path.exists(os.path.join(sessionName,"results")):
            os.makedirs(os.path.join(sessionName,"results"))

        VLMPsession = {"name":sessionName}
        VLMPsession["simulations"] = []
        VLMPsession["simulationSets"] = []
        for simSetIndex,simSet in enumerate(self.simulationSets):
            #Create folder sessionName/simulationSets/simulationSet_i
            simulationSetName   = f"simulationSet_{simSetIndex}"
            simulationSetFolder = os.path.join(sessionName,"simulationSets",simulationSetName)

            if not os.path.exists(simulationSetFolder):
                os.makedirs(simulationSetFolder)

            #For each simulation in the simulation set.
            #Create a folder sessionName/simulationSets/simulationSetName/simulationName/
            for simName in simSet:

                simulationFolder       = os.path.join(sessionName,"simulationSets",simulationSetName,simName)
                simulationResultFolder = os.path.join(sessionName,"results",simName)

                if not os.path.exists(simulationFolder):
                    os.makedirs(simulationFolder)

                if not os.path.islink(simulationResultFolder):
                    os.symlink(os.path.relpath(simulationFolder,
                                               "/".join(simulationResultFolder.split("/")[:-1])),
                               simulationResultFolder)

                #Update output files for each simulation in simSet
                sim = self.simulations[simName]

                #Write simulation file into results folder
                sim.write(os.path.join(simulationFolder,"simulation.json"))

                #Relative path to the simulation folder
                relativePath = os.path.relpath(simulationFolder,simulationSetFolder)

                VLMPsession["simulations"].append([simName,
                                                   os.path.join(*simulationFolder.split("/")[1:]),
                                                   os.path.join(*simulationResultFolder.split("/")[1:]),
                                                   self.simulationsInfo[simName]])

                #Updating file path
                def getValuesAndPaths(d, key, path=None):
                    """
                    Recursively search a nested dictionary
                    for all values associated with a given key,
                    along with the path to each value.
                    """
                    if path is None:
                        path = ()

                    values = []
                    for k, v in d.items():
                        new_path = path + (k,)
                        if k == key:
                            values.append((v, new_path))
                        elif isinstance(v, dict):
                            values.extend(getValuesAndPaths(v, key, new_path))

                    return values

                outputFilePaths = getValuesAndPaths(sim,"outputFilePath")
                for fName,fSimPath in outputFilePaths:
                    sim.setValue(fSimPath,os.path.join(relativePath,fName))

            ################################################
            #Aggregate simulations in simulation sets

            self.logger.debug(f"[VLMP] Aggregating simulations in simulation set {simSetIndex}")
            aggregatedSimulation = mergeSimulationsSet([self.simulations[simName] for simName in simSet])

            #Aggregated simulation is ready
            ################################################

            ################################################
            #Write aggregated simulation to file

            aggregatedSimulation.write(os.path.join(simulationSetFolder,f"simulationSet_{simSetIndex}.json"))

            #Relative path to the simulation folder
            relativePath = os.path.relpath(simulationSetFolder,sessionName)
            VLMPsession["simulationSets"].append([simulationSetName,
                                                  f"{relativePath}",
                                                  f"simulationSet_{simSetIndex}.json",
                                                  simSet.copy()])

        with open(os.path.join(sessionName,"VLMPsession.json"),"w") as simSetsFile:
            #Write simulation sets file using jsbeautifier
            simSetsFile.write(jsbeautifier.beautify(json.dumps(VLMPsession)))

        self.logger.debug("[VLMP] Simulation set up finished")



import sys, os
import itertools

import copy

import logging

from . import modelBase
from ...utils.geometry import getEx

import random

import orthopoly

import numpy as np
from scipy.spatial import cKDTree

from pyquaternion import Quaternion

from scipy.spatial.transform import Rotation as R
from scipy.optimize import root


class LEMON(modelBase):
    """
    Component name: LEMON
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 15/06/2023

    """

    def __generateHelix(self,yOffset=0.0,center=True,monomersOffset=0):

        sigma = 2.0*self.monomerRadius

        monomersPositions    = []
        monomersOrientations = []

        alpha = np.sqrt(self.theta0*self.theta0+self.phi0*self.phi0)

        cp = self.theta0/alpha;
        sp = self.phi0/alpha;

        cp2 = cp*cp;
        sp2 = sp*sp;

        ca = np.cos(alpha);
        sa = np.sin(alpha);

        Rtrans = np.asarray([[ca*cp2+sp2    ,-sa*cp ,(1.0-ca)*sp*cp],
                             [sa*cp         ,ca     ,-sa*sp],
                             [(1.0-ca)*sp*cp,sa*sp  ,ca*sp2+cp2]])

        r_trans = R.from_matrix(Rtrans)

        boxX,boxY,boxZ = [b/2.0 for b in self.box]

        for n in range(self.nMonomers+monomersOffset):
            if n > monomersOffset:
                self.logger.debug(f"[LEMON] Adding monomer {n}")

            if n==0:
                x=0.0
                y=-boxY+sigma+yOffset
                z=(self.theta0/(self.theta0*self.theta0+self.phi0*self.phi0)-1.0)

                Rinit = np.asarray([[ca*cp ,-sa ,-ca*sp],
                                    [sp    ,0.0 , cp],
                                    [-sa*cp,-ca , sa*sp]])

                r_init = R.from_matrix(Rinit)

                q1,q2,q3,q0 = r_init.as_quat()

            else:
                q0,q1,q2,q3 = monomersOrientations[-1]
                r_prev    = R.from_quat([q1,q2,q3,q0])
                r_current = r_prev*r_trans

                q1,q2,q3,q0 = r_current.as_quat()

                x,y,z = monomersPositions[-1]+0.5*sigma*(getEx(monomersOrientations[-1])+getEx([q0,q1,q2,q3]))


            currentMonomerPosition = np.asarray([x,y,z])
            currentMonomerOrientation = np.asarray([q0,q1,q2,q3])

            monomersPositions.append(currentMonomerPosition)
            monomersOrientations.append(currentMonomerOrientation)

        #Remove the first monomersOffset monomers
        monomersPositions    = monomersPositions[monomersOffset:]
        monomersOrientations = monomersOrientations[monomersOffset:]

        if(center):
            centroid = np.mean(monomersPositions,axis=0)
            monomersPositions-=centroid

        return np.asarray(monomersPositions),np.asarray(monomersOrientations)

    def __getMonomerConnections(self,position,orientation):
        connectionStart = position-self.monomerRadius*getEx(orientation)
        connectionEnd   = position+self.monomerRadius*getEx(orientation)

        return connectionStart,connectionEnd

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"mode","init",
                                                "nMonomers","box",
                                                "monomerRadius","patchRadius",
                                                "pitch","monomersPerTurn",
                                                "Eb","rc",
                                                "Kb","Ka","Kd",
                                                "stiffnessFactor",
                                                "epsilonLipids",
                                                "muLipids",
                                                "chiLipids",
                                                "thetaLipids"},
                         requiredParameters  = {"nMonomers","box",
                                                "pitch","monomersPerTurn",
                                                "Eb","rc",
                                                "Kb","Ka","Kd",
                                                "epsilonLipids",
                                                "muLipids",
                                                "chiLipids",
                                                "thetaLipids"},
                         definedSelections   = {"particleId"},
                         **params)

        self.logger = logging.getLogger("VLMP")

        ############################################################
        ####################  Model Parameters  ####################
        ############################################################

        self.kT = self.getUnits().getConstant("kT")

        self.init      = params.get("init","tube")

        self.nMonomers      = params["nMonomers"]

        self.box = params["box"]

        self.monomerRadius = params.get("monomerRadius",0.5)
        self.patchRadius   = params.get("patchRadius",0.1)

        self.pitch           = params["pitch"]
        self.monomersPerTurn = params["monomersPerTurn"]

        self.Eb = params["Eb"]
        self.rc = params["rc"]

        self.Kb = params["Kb"]
        self.Ka = params["Ka"]
        self.Kd = params["Kd"]

        stiffnessFactor = params.get("stiffnessFactor",1.0)
        if stiffnessFactor != 1.0:
            self.logger.info(f"[LEMON] Applying stiffness factor{stiffnessFactor}")
            self.Ka*=stiffnessFactor
            self.Kb*=stiffnessFactor

            self.logger.info(f"[LEMON] Kt theta after stiffness factor {self.Ka}")
            self.logger.info(f"[LEMON] Kt phi after stiffness factor {self.Kd}")

        def equationSystem(vars, nc, pitch, sigma):
            phi0, theta0 = vars

            a = pitch/(np.pi*2)
            b = nc*sigma/(np.pi*2)

            R = np.sqrt(b**2-a**2)

            alpha = np.sqrt(phi0**2+theta0**2)

            eq1 = phi0/alpha - pitch/(sigma*nc)
            eq2 = np.cos(alpha)*(theta0**2/alpha**2)+phi0**2/alpha**2-(a**2+np.cos(2*np.pi/nc)*(R**2))/(a**2+R**2)

            return [eq1, eq2]

        self.phi0,self.theta0 = root(equationSystem,
                                     [np.pi/2,np.pi/2],
                                     args=(self.monomersPerTurn,self.pitch,self.monomerRadius*2.0),tol=1e-12).x

        self.logger.info(f"[LEMON] Computing theta0 {self.theta0}")
        self.logger.info(f"[LEMON] Computing phi0 {self.phi0}")

        ##############################################

        #Lipids
        self.epsilonLipids = params["epsilonLipids"]
        self.muLipids      = params["muLipids"]
        self.chiLipids     = params["chiLipids"]
        self.thetaLipids   = params["thetaLipids"]

        self.logger.info(f"[LEMON] Epilson lipids {self.epsilonLipids}")
        self.logger.info(f"[LEMON] Mu lipids {self.muLipids}")
        self.logger.info(f"[LEMON] Chi lipids {self.chiLipids}")
        self.logger.info(f"[LEMON] Theta lipids {self.thetaLipids}")

        ##############################################

        self.logger.info(f"[LEMON] Generating LEMON model with {self.nMonomers} monomers")

        if self.init == "tube":
            nTotalMonomers = self.nMonomers #Backup

            monomersPerTube = [0 for i in range(self.pitch)]
            n=0
            i=0
            while n < self.nMonomers:
                if i >= self.pitch:
                    i=0
                monomersPerTube[i]+=1
                n+=1
                i+=1

            monomersPerOffset = int(self.monomersPerTurn/self.pitch)

            self.monomersPositions    = np.zeros((0,3))
            self.monomersOrientations = np.zeros((0,4))
            self.monomersHelixId      = np.zeros((0,1),dtype=int)
            for i in range(self.pitch):
                self.nMonomers = monomersPerTube[i]
                mPos,mOri = self.__generateHelix(yOffset=i*2.0*self.monomerRadius,center=False,monomersOffset=self.monomersPerTurn-monomersPerOffset*i)
                self.monomersPositions    = np.vstack((self.monomersPositions,mPos))
                self.monomersOrientations = np.vstack((self.monomersOrientations,mOri))
                #i is the helix id
                self.monomersHelixId      = np.vstack((self.monomersHelixId,np.ones((self.nMonomers,1),dtype=int)*i))


            centroid = np.mean(self.monomersPositions,axis=0)
            self.monomersPositions-=centroid

            self.nMonomers = nTotalMonomers #Restore
        else:
            self.logger.error(f"[LEMON] Init mode {self.init} is not avaible")
            raise Exception("Init mode not available")

        ##############################################

        types = self.getTypes()

        types.addType(name="M",radius=self.monomerRadius)

        ##############################################

        state = {}
        state["labels"]=["id","position","direction"]
        state["data"]  =[]

        for i,[p,q] in enumerate(zip(self.monomersPositions,self.monomersOrientations)):
            p=np.around(p,2)
            q=np.around(q,2)
            state["data"].append([int(i),list(p),list(q)])

        ##############################################

        structure = {}
        structure["labels"] = ["id", "type", "modelId"]
        structure["data"]   = []

        for i in range(self.nMonomers):
            #modelId is the helix id
            structure["data"].append([int(i),"M",int(self.monomersHelixId[i])])

        ##############################################

        forceField = {}

        #Verlet list
        forceField["verletListParticles"]={}
        forceField["verletListParticles"]["type"]       =  ["VerletConditionalListSet", "intra_inter"]
        forceField["verletListParticles"]["parameters"] =  {"cutOffVerletFactor":1.5}

        #WCA
        forceField["MonomerMonomerIntra"]={}
        forceField["MonomerMonomerIntra"]["type"]       =  ["NonBonded", "WCAType2"]
        forceField["MonomerMonomerIntra"]["parameters"] =  {"cutOffFactor":2.5,"condition":"intra"}
        forceField["MonomerMonomerIntra"]["labels"]     =  ["name_i", "name_j", "epsilon", "sigma"]
        forceField["MonomerMonomerIntra"]["data"]       =  [["M", "M", 1.0, 2.0*self.monomerRadius]]

        #ATZBERGER
        forceField["MonomerMonomerInter"] = {}
        forceField["MonomerMonomerInter"]["type"] = ["NonBonded", "Atzberger"]
        forceField["MonomerMonomerInter"]["parameters"] = {"condition":"inter","axis":[0.0,-1.0,0.0]}
        forceField["MonomerMonomerInter"]["labels"] = ["name_i", "name_j", "radius", "epsilon", "mu", "chi", "theta"]
        forceField["MonomerMonomerInter"]["data"]   = [
            ["M","M",self.monomerRadius,self.epsilonLipids,self.muLipids,self.chiLipids,self.thetaLipids]
        ]

        ############Patches

        forceField["lemon"]={}
        forceField["lemon"]["type"]       = ["PatchyParticles","DynamicallyBondedPatchyParticles"]

        forceField["lemon"]["patchesState"]={}
        forceField["lemon"]["patchesState"]["labels"]=["id","position"]
        forceField["lemon"]["patchesState"]["data"]  =[]

        for i in range(self.nMonomers):

            index=i*2
            p = [-self.monomerRadius,0.0,0.0]
            forceField["lemon"]["patchesState"]["data"].append([int(index  ),list(p)])
            p = [ self.monomerRadius,0.0,0.0]
            forceField["lemon"]["patchesState"]["data"].append([int(index+1),list(p)])

        forceField["lemon"]["patchesGlobal"]={}
        forceField["lemon"]["patchesGlobal"]["fundamental"] = {}
        forceField["lemon"]["patchesGlobal"]["fundamental"]["type"]      = ["Fundamental","DynamicallyBondedPatchyParticles"]
        forceField["lemon"]["patchesGlobal"]["fundamental"]["parameters"]={"energyThreshold":0.0}

        forceField["lemon"]["patchesGlobal"]["types"]  = {}
        forceField["lemon"]["patchesGlobal"]["types"]["type"]   = ["Types","Basic"]
        forceField["lemon"]["patchesGlobal"]["types"]["labels"] = ["name", "mass", "radius", "charge"]
        forceField["lemon"]["patchesGlobal"]["types"]["data"]   = [["S",0.0,self.patchRadius,0.0],
                                                                      ["E",0.0,self.patchRadius,0.0]]

        forceField["lemon"]["patchesTopology"]={}

        forceField["lemon"]["patchesTopology"]["structure"]={}
        forceField["lemon"]["patchesTopology"]["structure"]["labels"] = ["id", "type", "parentId"]
        forceField["lemon"]["patchesTopology"]["structure"]["data"]   = []

        for i in range(self.nMonomers):
            index=i*2
            forceField["lemon"]["patchesTopology"]["structure"]["data"].append([index  ,"E",i])
            forceField["lemon"]["patchesTopology"]["structure"]["data"].append([index+1,"S",i])

        forceField["lemon"]["patchesTopology"]["forceField"] = {}

        #Verlet list
        forceField["lemon"]["patchesTopology"]["forceField"]["verletList"]={}
        forceField["lemon"]["patchesTopology"]["forceField"]["verletList"]["type"]       =  ["VerletConditionalListSet", "interDifferentType"]
        forceField["lemon"]["patchesTopology"]["forceField"]["verletList"]["parameters"] =  {"cutOffVerletFactor":3.0}

        forceField["lemon"]["patchesTopology"]["forceField"]["transversal"]={}
        forceField["lemon"]["patchesTopology"]["forceField"]["transversal"]["type"]       =  ["NonBondedPatches", "Helix"]
        forceField["lemon"]["patchesTopology"]["forceField"]["transversal"]["parameters"] =  {"condition":"inter"}

        forceField["lemon"]["patchesTopology"]["forceField"]["transversal"]["labels"]     =  ["name_i", "name_j", "Eb", "Kb", "Ka", "Kd", "rc", "theta0", "phi0"]
        forceField["lemon"]["patchesTopology"]["forceField"]["transversal"]["data"]       =  [["S", "E",
                                                                                                  self.Eb,
                                                                                                  self.Kb, self.Ka, self.Kd,
                                                                                                  self.rc,
                                                                                                  self.theta0, self.phi0]]

        ##############################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)

    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        return sel



from VLMP.components.modelExtensions import modelExtensionBase

class AFM(modelExtensionBase):

    """
    Component name: AFM
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    AFM model extension.

    :param epsilon: epsilon parameter for tip-particle interaction
    :type epsilon: float
    :param sigma: sigma parameter for tip-particle interaction
    :type sigma: float
    :param K: spring constant
    :type K: float
    :param Kxy: xy spring constant
    :type Kxy: float
    :param tipVelocity: tip velocity
    :type tipVelocity: float
    :param startChipPosition: initial position of the chip
    :type startChipPosition: list of float [x,y,z]
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"epsilon",
                                                "sigma",
                                                "K","Kxy",
                                                "tipVelocity",
                                                "indentationStartStep",
                                                "indentationBackwardStep"},
                         requiredParameters  = {"epsilon",
                                                "sigma",
                                                "K","Kxy",
                                                "tipVelocity"},
                         availableSelections = {"tip","sample"},
                         requiredSelections  = {"tip","sample"},
                         **params)

        epsilon = params["epsilon"]
        sigma   = params["sigma"]

        K   = params["K"]
        Kxy = params["Kxy"]

        tipVelocity = params["tipVelocity"]

        indentationStartStep    = params.get("indentationStartStep",0)
        indentationBackwardStep = params.get("indentationBackwardStep",0)

        ############################################################

        tipIds    = self.getSelection("tip")
        sampleIds = self.getSelection("sample")

        # Check tipIds has only one element
        if len(tipIds) != 1:
            self.logger.error("AFM model extension only supports tips made of one particle")
            raise Exception("Not supported tip selection")
        else:
            tipPos = self.getIdsState(tipIds,"position")
            startChipPosition = tipPos[0]

            self.logger.debug(f"AFM model extension, startChipPosition: {startChipPosition}")


        extension = {}

        extension[name] = {}
        extension[name]["type"]       = ["AFM","SphericalTip"]
        extension[name]["parameters"] = {}

        extension[name]["labels"] = ["idSet_i","idSet_j","epsilon","sigma","K","Kxy","tipVelocity","startChipPosition","indentationStartStep","indentationBackwardStep"]
        extension[name]["data"]   = [[tipIds,sampleIds,epsilon,sigma,K,Kxy,tipVelocity,startChipPosition,indentationStartStep,indentationBackwardStep]]

        ############################################################

        self.setExtension(extension)



def picosecond2KcalMol_A_time(picosecond=1.0):
    return (20/0.978)*picosecond

def KcalMol_A_time2picosecond(KcalMol_A_time=1.0):
    return (0.978/20)*KcalMol_A_time

def KcalMol_A_force2nanonewton(KcalMol_div_A=1.0):
    return 0.06948*KcalMol_div_A

def nanonewton2KcalMol_A_force(nanonewton=1.0):
    return (1.0/0.06948)*nanonewton



import sys, os

import logging

import numpy as np

from . import modelExtensionBase
from ...utils.input import getLabelIndex

class intraSteric(modelExtensionBase):

    """
    Component name: intraSteric
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 21/09/2023

    Steric interactions between atoms of the same molecule. If molecues are bonded, the interaction is not considered.

    :param epsilon: epsilon parameter for the interaction
    :type epsilon: float
    :param cutOffFactor: Factor to multiply the sigma parameter to obtain the cut-off distance.
    :type cutOffFactor: float
    :param addVerletList: If True, a Verlet list will be created for the interactions.
    :type addVerletList: bool, optional, default=True

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"epsilon","cutOffFactor","excludedBonds","addVerletList"},
                         requiredParameters  = {"epsilon","cutOffFactor"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        epsilon            = params.get("epsilon")
        cutOffFactor       = params.get("cutOffFactor",1.0)

        excludedBonds      = params.get("excludedBonds",0)
        addVerletList      = params.get("addVerletList",True)

        if excludedBonds > 0 and addVerletList == False:
            self.logger.error("[intraSteric] excludedBonds > 0 and addVerletList == False. This is not allowed. Exiting...")
            raise Exception("Not compatible parameters.")

        ############################################################

        # Build the interaction matrix
        interactionMatrix = []

        types = self.getTypes().getTypes()
        for t1 in types:
            for t2 in types:
                interaction = [t1,t2,epsilon,types[t1]["radius"] + types[t2]["radius"]]
                interactionMatrix.append(interaction)

        extension = {}

        if addVerletList:
            extension["nl"]={}
            extension["nl"]["type"]       =  ["VerletConditionalListSet", "nonExclIntra_nonExclInter"]
            extension["nl"]["parameters"] =  {}

            exclusions = {}

            if excludedBonds > 0:
                #Add bond2
                for mdl in self.getModels():
                    for bondName,info in mdl.getForceField().items():
                        type_ = info["type"][0]
                        if type_ == "Bond2":
                            index_i = getLabelIndex("id_i",info["labels"])
                            index_j = getLabelIndex("id_j",info["labels"])

                            for d in info["data"]:
                                i = d[index_i]+mdl.getIdOffset()
                                j = d[index_j]+mdl.getIdOffset()

                                if i not in exclusions:
                                    exclusions[i] = set()
                                if j not in exclusions:
                                    exclusions[j] = set()

                                exclusions[i].add(j)
                                exclusions[j].add(i)

            if excludedBonds > 1:
                #Add bond3
                for mdl in self.getModels():
                    for bondName,info in mdl.getForceField().items():
                        type_ = info["type"][0]
                        if type_ == "Bond3":
                            index_i = getLabelIndex("id_i",info["labels"])
                            index_j = getLabelIndex("id_j",info["labels"])
                            index_k = getLabelIndex("id_k",info["labels"])

                            for d in info["data"]:
                                i = d[index_i]+mdl.getIdOffset()
                                j = d[index_j]+mdl.getIdOffset()
                                k = d[index_k]+mdl.getIdOffset()

                                if i not in exclusions:
                                    exclusions[i] = set()
                                if j not in exclusions:
                                    exclusions[j] = set()
                                if k not in exclusions:
                                    exclusions[k] = set()

                                exclusions[i].add(j)
                                exclusions[i].add(k)
                                exclusions[j].add(i)
                                exclusions[j].add(k)
                                exclusions[k].add(i)
                                exclusions[k].add(j)

            if excludedBonds > 2:
                #Add bond4
                for mdl in self.getModels():
                    for bondName,info in mdl.getForceField().items():
                        type_ = info["type"][0]
                        if type_ == "Bond4":
                            index_i = getLabelIndex("id_i",info["labels"])
                            index_j = getLabelIndex("id_j",info["labels"])
                            index_k = getLabelIndex("id_k",info["labels"])
                            index_l = getLabelIndex("id_l",info["labels"])

                            for d in info["data"]:
                                i = d[index_i]+mdl.getIdOffset()
                                j = d[index_j]+mdl.getIdOffset()
                                k = d[index_k]+mdl.getIdOffset()
                                l = d[index_l]+mdl.getIdOffset()

                                if i not in exclusions:
                                    exclusions[i] = set()
                                if j not in exclusions:
                                    exclusions[j] = set()
                                if k not in exclusions:
                                    exclusions[k] = set()
                                if l not in exclusions:
                                    exclusions[l] = set()

                                exclusions[i].add(j)
                                exclusions[i].add(k)
                                exclusions[i].add(l)
                                exclusions[j].add(i)
                                exclusions[j].add(k)
                                exclusions[j].add(l)
                                exclusions[k].add(i)
                                exclusions[k].add(j)
                                exclusions[k].add(l)
                                exclusions[l].add(i)
                                exclusions[l].add(j)
                                exclusions[l].add(k)

            extension["nl"]["labels"]      = ["id","id_list"]
            extension["nl"]["data"]        = []
            for i in exclusions:
                extension["nl"]["data"].append([i,list(exclusions[i])])

        extension[name] = {}
        extension[name]["type"] = ["NonBonded","WCAType2"]
        extension[name]["parameters"] = {"cutOffFactor":cutOffFactor,"condition":"intra"}
        extension[name]["labels"] = ["name_i","name_j","epsilon","sigma"]
        extension[name]["data"]   = interactionMatrix.copy()

        ############################################################

        self.setExtension(extension)






import sys, os
import copy

import logging

import numpy as np

from . import modelExtensionBase

class constraintCenterOfMassPosition(modelExtensionBase):

    """
    Component name: constraintCenterOfMassPosition
    Component type: modelExtension

    Author: Pablo Ibáñez-Freire
    Date: 14/03/2023

    Apply a constraint to the center of mass of a selection of particles

    :param selection: Selection of particles where the constraint will be applied
    :type selection: list of dictionaries
    :param K: Stiffness of the constraint
    :type K: float
    :param r0: Distance between the center of mass and the constraint position
    :type r0: float
    :param position: Position of the center of mass of the selection
    :type position: list of floats

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"K","r0","position"},
                         requiredParameters  = {"K","r0","position"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        K        = params.get("K")
        #Check if K is a float
        if not isinstance(K,float) and not isinstance(K,list):
            raise Exception("K must be a float or a list of floats")
        if isinstance(K,float):
            K = [K,K,K]

        r0       = params.get("r0")
        #Check if r0 is a float
        if not isinstance(r0,float):
            raise Exception("r0 must be a float")

        position = params.get("position")
        #Check if position is a list of floats
        if not isinstance(position,list):
            raise Exception("position must be a list of floats")

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["Set1","FixedHarmonicCenterOfMass"]
        extension[name]["parameters"] = {}
        extension[name]["labels"] = ["idSet_i","K","r0","position"]
        extension[name]["data"]   = []

        selectedIds = self.getSelection("selection")

        extension[name]["data"].append([selectedIds,copy.deepcopy(K),[r0,r0,r0],position])

        ############################################################

        self.setExtension(extension)






import sys, os

import logging

from . import systemBase

class simulationName(systemBase):
    """
    Component name: simulationName
    Component type: system

    Author: Pablo Ibáñez-Freire
    Date: 13/03/2023

    Simple component to add a name to the simulation.

    :param simulationName: Name of the simulation
    :type simulationName: str
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"simulationName"},
                         requiredParameters  = {"simulationName"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        system = {
            name:{"type":["Simulation","Information"],
                  "parameters":{}}
        }

        system[name]["parameters"]["name"] = params.get("simulationName")

        ############################################################

        self.setSystem(system)




import logging

from MDAnalysis.lib.pkdtree import PeriodicKDTree

import numpy as np

def distributeRandomlyGeneratorChecker(box,posSets,radSets,newPosGenerator,distanceChecker,nMaxTries,radiusFactor):

    logger = logging.getLogger("VLMP")

    newPositions = []
    addedRads    = []

    maxRadius = np.max([np.max(r) for r in radSets])

    for i in range(len(posSets)):
        #Trying to find a new position for the particle set i

        if len(newPositions) == 0:
            center = np.mean(posSets[i],axis=0)
            # Generete a random 3D position
            mp = newPosGenerator()

            transVec = mp - center
            newPositions += [list(p + transVec) for p in posSets[i]]
            addedRads+=radSets[i]
            logger.debug(f"Added particle 1/{len(posSets)}")
            continue

        tree = PeriodicKDTree(box=np.asarray(box+[90,90,90],dtype=np.float32))
        tree.set_coords(coords=np.asarray(newPositions,dtype=np.float32),
                        cutoff=maxRadius*2.0)

        # Try to add the set in a random position
        added = False
        tries = 0
        while not added and tries < nMaxTries:
            center = np.mean(posSets[i],axis=0)

            # Generate a random position
            mp = newPosGenerator()

            transVec = mp - center

            tentativePos = np.asarray([p + transVec for p in posSets[i]])

            # Find the closest particles
            pairs = tree.search_tree(np.asarray(tentativePos,dtype=np.float32),maxRadius*2.0)

            toAdd = True
            for i1,i2 in pairs:

                p1 = newPositions[i2]
                p2 = tentativePos[i1]

                r1 = addedRads[i2]
                r2 = radSets[i][i1]

                toAdd = distanceChecker(p1,p2,r1,r2)
                if not toAdd:
                    break
            if toAdd:
                newPositions += [list(p) for p in tentativePos]
                addedRads    += radSets[i]
                added = True
                logger.debug(f"Added particle {i+1}/{len(posSets)} at try {tries}/{nMaxTries}")
            else:
                tries += 1

        if not added:
            logger.error("The number of tries to avoid clashes has been reached.")
            raise Exception("Clash avoidance failed")

    return newPositions.copy()




import logging

########################################################

from ..utils.input import getLabelIndex

class idsHandler:

    _models     = None

    _id2model   = None
    _id2localId = None

    def __getModelLocalId(self, i):
        return idsHandler._id2model[i], idsHandler._id2localId[i]

    def __init__(self,
                 models):

        logger = logging.getLogger("VLMP")

        if idsHandler._models is None or idsHandler._models != models:

            logger.debug("Initializing idsHandler")

            idsHandler._models = models

            N = sum([mdl.getNumberOfParticles() for mdl in idsHandler._models])

            idsHandler._id2model   = [None]*N
            idsHandler._id2localId = [None]*N

            offset = 0
            for mdlIndex,mdl in enumerate(idsHandler._models):
                for i in range(mdl.getNumberOfParticles()):
                    idsHandler._id2model[i+offset]   = mdlIndex
                    idsHandler._id2localId[i+offset] = i

                offset += mdl.getNumberOfParticles()

            logger.debug("Done initializing idsHandler")
        else:
            logger.debug("idsHandler already initialized")

    ######################## GETTERS #######################

    def _getIdsProperty(self,globalIds,propertyName):
        idsProperty = []

        for i in globalIds:
            mdlIndex, localId = self.__getModelLocalId(i)

            mdl = idsHandler._models[mdlIndex]

            typeIndex = getLabelIndex("type",mdl.getStructure()["labels"])
            itype     = mdl.getStructure()["data"][localId][typeIndex]

            idsProperty.append(mdl.getTypes().getTypes()[itype][propertyName])

        return idsProperty

    def _getIdsState(self,globalIds,stateName):
        idsState = []

        for i in globalIds:
            mdlIndex, localId = self.__getModelLocalId(i)

            mdl = idsHandler._models[mdlIndex]

            stateIndex = getLabelIndex(stateName,mdl.getState()["labels"])
            stateValue = mdl.getState()["data"][localId][stateIndex]

            idsState.append(stateValue)

        return idsState

    def _getIdsStructure(self,globalIds,structureName):

        totalParticles = sum([mdl.getNumberOfParticles() for mdl in idsHandler._models])

        id2struct = [None]*totalParticles
        structOffset = 0
        for mdl in idsHandler._models:
            structLabels = mdl.getStructure()["labels"]
            for i in mdl.getGlobalIds():
                mdlIndex, localId = self.__getModelLocalId(i)
                mdl = idsHandler._models[mdlIndex]

                maxStruct = structOffset
                if structureName in structLabels:
                    structIndex  = getLabelIndex(structureName,structLabels)
                    structValue  = mdl.getStructure()["data"][localId][structIndex]

                    id2struct[i] = structValue+structOffset
                else:
                    id2struct[i] = structOffset

                maxStruct = max(maxStruct,id2struct[i])

            structOffset += maxStruct+1

        ########################################################

        idsStructure = []

        for i in globalIds:
            idsStructure.append(id2struct[i])

        return idsStructure

    ######################## SETTERS #######################

    def _setIdsState(self,globalIds,stateName,states):

        logger = logging.getLogger("VLMP")

        if len(globalIds) != len(states):
            logger.error(f"[ModelOperation] Number of ids and states ({stateName}) do not match")
            raise Exception(f"Number of ids and states do not match")

        for i,s in zip(globalIds,states):
            mdlIndex, localId = self.__getModelLocalId(i)
            mdl = idsHandler._models[mdlIndex]

            stateIndex = getLabelIndex(stateName,mdl.getState()["labels"])

            #Check if state is valid
            if type(s) != type(mdl.getState()["data"][localId][stateIndex]):
                logger.error(f"[ModelOperation] State value {s} for state \"{stateName}\" is not valid."
                             f" State value type is {type(s)} but should be {type(mdl.getState()['data'][localId][stateIndex])}")
                raise Exception(f"State is not valid")
            else:
                if type(s) == list:
                    if len(s) != len(mdl.getState()["data"][localId][stateIndex]):
                        logger.error(f"[ModelOperation] State value {s} for state \"{stateName}\" is not valid, length does not match")
                        raise Exception(f"State is not valid")

                mdl.getState()["data"][localId][stateIndex] = s



import sys, os

import logging

import numpy as np

from . import modelExtensionBase

class ACMagneticField(modelExtensionBase):

    """
    Component name: ACMagneticField
    Component type: modelExtension

    Author: P. Palacios-Alonso
    Date: 16/10/2023

    Alternating current (AC) magnetic field applied to a selection of magnetic particles

    :param selection: Selection of particles where the force is applied
    :type selection: list of dictionaries
    :param force: Force applied to the particles
    :type force: list of floats

    ...
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"b0", "frequency", "direction"},
                         requiredParameters  = {"b0", "frequency"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        frequency = params["frequency"]
        b0        = params["b0"]
        direction = params.get("direction", [0,0,1])

        extension = {}

        extension[name] = {}
        extension[name]["type"] = ["External","ACMagneticField"]
        extension[name]["parameters"] = {}
        extension[name]["parameters"]["b0"] = b0
        extension[name]["parameters"]["frequency"] = frequency
        extension[name]["parameters"]["direction"] = direction.copy()
        ############################################################

        self.setExtension(extension)






import sys,os

import logging

import json

################### DEBUG MODE ##################

DEBUG_MODE = True

################# SET UP LOGGER #################

class CustomFormatter(logging.Formatter):

    white = "\x1b[37;20m"
    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format     = "%(asctime)s - %(name)s - %(levelname)s - %(message)s "
    formatLine = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: white + formatLine + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + formatLine + reset,
        logging.CRITICAL: bold_red + formatLine + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt,datefmt='%d/%m/%Y %H:%M:%S')
        return formatter.format(record)

logger = logging.getLogger("VLMP")
logger.handlers = []
logger.setLevel(logging.DEBUG)

clogger = logging.StreamHandler()
if DEBUG_MODE:
    clogger.setLevel(logging.DEBUG) #<----
else:
    clogger.setLevel(logging.INFO) #<----

clogger.setFormatter(CustomFormatter())
logger.addHandler(clogger)

#################################################

if "-m" not in sys.argv:
    #This done to avoid the error when running the program with -m option.
    #When running with -m option a large number of libraries imported by VLMP
    #are not needed.

    logger.info("Starting VLMP...")

    from .VLMP  import VLMP



import VLMP

import VLMP.components.units as _units

import os

import numpy as np
import matplotlib.pyplot as plt

import json
import jsbeautifier

import copy
import logging

class SurfaceUmbrellaSampling(VLMP.VLMP):

    def __init__(self,parameters):
        super().__init__()

        requiredParameters = ["nWindows","windowsStartPosition","windowsEndPosition","K","Ksteps"]

        self.logger.info("[SurfaceUmbrellaSampling] Initializing ...")

        #Load umbrella parameters

        for parameter in requiredParameters:
            if parameter not in parameters["umbrella"]:
                self.logger.error("[SurfaceUmbrellaSampling] Required parameter %s not found!" % parameter)
                raise Exception("Required parameter not found!")

        self.nWindows        = parameters["umbrella"]["nWindows"]

        self.windowsStartPosition    = parameters["umbrella"]["windowsStartPosition"]
        self.windowsEndPosition      = parameters["umbrella"]["windowsEndPosition"]

        self.windowsSize             = (self.windowsEndPosition - self.windowsStartPosition)/(self.nWindows-1)

        self.K                       = parameters["umbrella"]["K"]
        self.Ksteps                  = parameters["umbrella"]["Ksteps"]
        if type(self.K) is not list:
            self.K = [self.K]

        if type(self.Ksteps) is not list:
            self.Ksteps = [self.Ksteps]

        if len(self.K) != len(self.Ksteps):
            self.logger.error("[SurfaceUmbrellaSampling] K and Ksteps must have the same length!")
            raise Exception("K and Ksteps must have the same length!")

        #Measurements
        self.measurementsIntervalStep = parameters["umbrella"]["measurementsIntervalStep"]

        self.logger.info((f"[SurfaceUmbrellaSampling] Number of windows: {self.nWindows}, "
                          f"windows start position: {self.windowsStartPosition}, "
                          f"windows end position: {self.windowsEndPosition}, "
                          f"windows size: {round(self.windowsSize,2)}",
                          f"measurements interval step: {self.measurementsIntervalStep}"))

        self.windowPositions = []
        for i in range(self.nWindows):
            self.windowPositions.append(self.windowsStartPosition + i*self.windowsSize)

        #Load simulation parameters

        requiredSimulationParameters = ["units","types","temperature","box","models","selection","integrator"]
        for parameter in requiredSimulationParameters:
            if parameter not in parameters["simulation"]:
                self.logger.error("[SurfaceUmbrellaSampling] Required parameter %s not found!" % parameter)
                raise Exception("Required parameter not found!")

        self.units  = parameters["simulation"]["units"]
        self.types  = parameters["simulation"]["types"]

        self.temperature = parameters["simulation"]["temperature"]
        #Check if the temperature is a float
        if not isinstance(self.temperature,float):
            raise Exception("The temperature must be a float")

        self.box         = parameters["simulation"]["box"]
        #Check box is a list of 3 floats
        if not isinstance(self.box,list):
            raise Exception("The box must be a list of 3 floats")
        if len(self.box) != 3:
            raise Exception("The box must be a list of 3 floats")
        for i in self.box:
            if not isinstance(i,float):
                raise Exception("The box must be a list of 3 floats")

        self.integrator = parameters["simulation"]["integrator"]
        self.integrator["parameters"]["integrationSteps"] = sum(self.Ksteps)

        self.models     = parameters["simulation"]["models"]
        self.selection  = parameters["simulation"]["selection"]

        self.backupIntervalStep = parameters["simulation"].get("backupIntervalStep",None)

        requiredOutputParameters = []
        for parameter in requiredOutputParameters:
            if parameter not in parameters["output"]:
                self.logger.error("[SurfaceUmbrellaSampling] Required parameter %s not found!" % parameter)
                raise Exception("Required parameter not found!")

        #Load output parameters

        #Info
        self.infoIntervalStep = parameters["output"].get("infoIntervalStep",None)

        #Save state
        self.saveStateIntervalStep   = parameters["output"].get("saveStateIntervalStep",None)
        self.saveStateOutputFilePath = parameters["output"].get("saveStateOutputFilePath",None)
        self.saveStateOutputFormat   = parameters["output"].get("saveStateOutputFormat",None)

        #If at least one of the save state parameters is specified, check all of them are specified
        self.saveState = False
        if self.saveStateIntervalStep is not None or self.saveStateOutputFilePath is not None or self.saveStateOutputFormat is not None:
            self.saveState = True
            if self.saveStateIntervalStep is None or self.saveStateOutputFilePath is None or self.saveStateOutputFormat is None:
                self.logger.error("[SurfaceUmbrellaSampling] All the save state parameters must be specified")
                raise Exception("All the save state parameters must be specified")

    def generateSimulationPool(self):

        self.umbrellaInfo = {}

        ########################################

        unitsComponent = eval(f"_units.{self.units}")(name="units")

        ########################################

        simulationPool = []

        mdlNames = []
        for mdl in self.models:
            if "name" not in mdl.keys():
                self.logger.error("[SurfaceUmbrellaSampling] Model name not found! (Model: %s)" % mdl)
                raise Exception("Model name not found")

            if mdl["name"] in mdlNames:
                self.logger.error("[SurfaceUmbrellaSampling] Model with the same name found! (Model: %s)" % mdl)
                raise Exception("Model with the same name found")
            else:
                mdlNames.append(mdl["name"])

        for mdl in self.models:
            self.umbrellaInfo[mdl["name"]] = {}
            self.umbrellaInfo[mdl["name"]]["kT"] = unitsComponent.getConstant("KBOLTZ")*self.temperature
            self.umbrellaInfo[mdl["name"]]["K"]  = self.K[-1]
            self.umbrellaInfo[mdl["name"]]["centers"] = {}
            for i,center in enumerate(self.windowPositions):

                sim = {"system":[{"type":"simulationName","parameters":{"simulationName":mdl["name"]+"_"+str(i)}}],
                       "units":[{"type":self.units}],
                       "types":[{"type":self.types}],
                       "ensemble":[{"type":"NVT","parameters":{"box":self.box,"temperature":self.temperature}}],
                       "integrators":[self.integrator],
                       "models":[mdl],
                       "modelExtensions":[],
                       "simulationSteps":[]

                       }

                for ik,k in enumerate(self.K):
                    sim["modelExtensions"].append({"name":f"constraint_{ik}",
                                                   "type":"constraintCenterOfMassPosition",
                                                   "parameters":{"K":k,
                                                                 "r0":0.0,
                                                                 "position":[0.0,0.0,center],
                                                                 "selection":{"expression":self.selection}
                                                                }
                                                   })
                if len(self.K) > 1:
                    stepsSum = 0
                    for ik,steps in enumerate(self.Ksteps):
                        sim["modelExtensions"][ik]["parameters"]["startStep"] = stepsSum
                        if ik != len(self.Ksteps)-1:
                            sim["modelExtensions"][ik]["parameters"]["endStep"]   = stepsSum + steps
                        stepsSum += steps

                #Add measures
                measureStartStep = 0
                #Iterate over all Ksteps but the last one
                for ik,steps in enumerate(self.Ksteps[:-1]):
                    measureStartStep += steps

                centerOutputFilePath = f"constraint_{center}.dat"
                sim["simulationSteps"].append({"type":"centerOfMassMeasurement","parameters":{"outputFilePath":centerOutputFilePath,
                                                                                              "intervalStep":self.measurementsIntervalStep,
                                                                                              "startStep":measureStartStep,
                                                                                              "selection":{"expression":self.selection}}})

                self.umbrellaInfo[mdl["name"]]["centers"][center] = "results/{}/{}".format(mdl["name"]+"_"+str(i),centerOutputFilePath)

                #Add backup
                if self.backupIntervalStep is not None:
                    sim["system"].append({"type":"backup","parameters":{"backupIntervalStep":self.backupIntervalStep}})

                #Add info
                if self.infoIntervalStep is not None:
                    sim["simulationSteps"].append({"type":"info","parameters":{"intervalStep":self.infoIntervalStep}})

                #Add save state
                if self.saveState:
                    sim["simulationSteps"].append({"type":"saveState","parameters":{"intervalStep":self.saveStateIntervalStep,
                                                                                    "outputFilePath":self.saveStateOutputFilePath,
                                                                                    "outputFormat":self.saveStateOutputFormat}})

                simulationPool.append(sim.copy())

        self.loadSimulationPool(copy.deepcopy(simulationPool))

    def setUpSimulation(self, sessionName):
        super().setUpSimulation(sessionName)

        #Write umbrella info
        with open(os.path.join(sessionName,"surfaceUmbrella.json"),"w") as f:
            f.write(jsbeautifier.beautify(json.dumps(self.umbrellaInfo)))

class AnalysisSurfaceUmbrellaSampling:

    def __computePotential(self,centers,data,beta,K,outputFolderPath):

        outputPlot      = os.path.join(outputFolderPath,"histogram.png")
        outputPotential = os.path.join(outputFolderPath,"potential.dat")

        #Check if histogram.png and potential.dat already exist
        if os.path.exists(outputPlot) and os.path.exists(outputPotential):
            self.logger.info("[AnalysisSurfaceUmbrellaSampling] Histogram and potential already computed")
            return True

        from WHAM import binless
        from WHAM.lib import potentials

        fig, (ax1, ax2) = plt.subplots(2,1)

        potMin = np.amin(np.array([np.amin(np.array(d)) for d in data]).flatten())
        potMax = np.amax(np.array([np.amin(np.array(d)) for d in data]).flatten())

        ######### HISTOGRAM ########

        cmap = plt.get_cmap('spring')
        ax1.set_prop_cycle(color=[cmap(1.*i/len(centers)) for i in range(len(centers))])

        for i,x_0 in enumerate(centers):
            traj = data[i]
            bins = 100
            ax1.hist(traj, bins=bins, alpha=0.8,density=True)

        ax1.set_xlabel("z")
        ax1.set_xlim(potMin, potMax)
        ax1.set_ylabel("Counts")

        ######### POTENTIAL ########

        x_it = []
        u_i  = []
        for i,x_0 in enumerate(centers):
            traj = data[i]
            x_it.append(traj)
            u_i.append(potentials.harmonic(K, x_0))

        #Check all traj in x_it have the same length
        if not all(len(x_it[0]) == len(x) for x in x_it):
            if self.ignoreDifferentLength:
                self.logger.error("[AnalysisSurfaceUmbrellaSampling] Not all trajectories have the same length. Ignoring ...")
                return False
            else:
                self.logger.warning("[AnalysisSurfaceUmbrellaSampling] Not all trajectories have the same length.")

        #Smamllest center -> potMin
        #Largest center  -> potMax
        #potMin = np.amin(centers)
        potMax = np.amax(centers)
        x_bin  = np.linspace(potMin,potMax,1001)

        calc_binless = binless.Calc1D()
        bF, _, _ = calc_binless.compute_betaF_profile(x_it, x_bin, u_i, beta=beta)

        bF = bF - np.min(bF)

        ax2.plot(x_bin, bF)

        ax2.set_xlabel("z")
        ax2.set_xlim(potMin, potMax)
        ax2.set_ylabel(r"Free energy ($k_B T$)")

        with open(outputPotential,"w") as fpot:
            for x,e in zip(x_bin, bF):
                fpot.write(f"{x} {e}\n")

        fig.set_size_inches(24, 17)
        fig.savefig(outputPlot, dpi=300)

        return True

    def __init__(self,
                 infoFilePath,
                 skip=0,
                 ignoreDifferentLength=True):

        self.logger = logging.getLogger("VLMP")

        with open(infoFilePath,"r") as f:
            self.info = json.load(f)
        #Path from current file to infoFilePath
        self.sessionPath = os.path.dirname(infoFilePath)

        self.skip = skip + 1
        self.ignoreDifferentLength = ignoreDifferentLength

        ########################################################

    def run(self):

        for mdlName in self.info.keys():

            self.logger.info(f"[AnalysisSurfaceUmbrellaSampling] Processing {mdlName}")

            kT   = self.info[mdlName]["kT"]
            beta = 1.0/kT

            K = self.info[mdlName]["K"]

            centers = []
            data    = []

            process = True
            for center,file in self.info[mdlName]["centers"].items():
                #Check if file exists
                if not os.path.isfile(os.path.join(self.sessionPath,file)):
                    self.logger.error(f"[AnalysisSurfaceUmbrellaSampling] Center file not found! (File: {file})")
                    process = False
                    break

                try:
                    centers.append(float(center))
                    data.append(np.loadtxt(os.path.join(self.sessionPath,file),skiprows=self.skip)[:,3])
                except:
                    self.logger.error(f"[AnalysisSurfaceUmbrellaSampling] Error loading file {file}")
                    process = False
                    break

            if not process:
                continue

            outputFolderPath = os.path.join(self.sessionPath,"results",mdlName+"_surfaceUmbrella")
            #Create output directory if it does not exist
            if not os.path.isdir(outputFolderPath):
                os.makedirs(outputFolderPath)

            if(self.__computePotential(centers,data,beta,K,outputFolderPath)):
                self.logger.info(f"[AnalysisSurfaceUmbrellaSampling] Results for model {mdlName} saved in {outputFolderPath}")
            else:
                self.logger.error(f"[AnalysisSurfaceUmbrellaSampling] Error while computing potential for model {mdlName}")






import sys, os

import logging

from . import simulationStepBase

class info(simulationStepBase):
    """
    Component name: saveState
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 13/03/2023

    Simple info step, it shows the current step,
    an estimation of the remaining time and the mean FPS.

    :param startStep: first step to show the info
    :type startStep: int, optional
    :param endStep: last step to show the info
    :type endStep: int, optional

    """

    def __init__(self,name,**params):
        super().__init__(_type= self.__class__.__name__,
                         _name= name,
                         availableParameters = set(),
                         requiredParameters  = set(),
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        simulationStep = {
            name:{
                  "type":["UtilsStep","InfoStep"],
                  "parameters":{}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)





from VLMP.components.modelOperations import modelOperationBase

import numpy as np

from scipy.spatial.transform import Rotation as R

class alignInertiaMomentAlongVector(modelOperationBase):

    """
    Component name: alignInertiaMomentAlongVector
    Component type: modelOperation

    Author: Pablo Ibáñez-Freire
    Date: 17/06/2023

    Aling the largest inertia moment of the selected elements along a given vector.

    :param vector: Vector along which the inertia moment will be aligned.
    :type vector: list of floats

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"vector"},
                         requiredParameters  = {"vector"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        vector = np.asarray(params["vector"])
        vector /= np.linalg.norm(vector)

        ############################################################

        selectedIds = self.getSelection("selection")

        if len(selectedIds) == 0:
            self.error("No elements selected.")
            raise Exception("No elements selected.")

        if len(selectedIds) > 1:

            masses = np.asarray(self.getIdsProperty(selectedIds,"mass"))
            pos    = np.asarray(self.getIdsState(selectedIds,"position"))

            inertia = np.zeros((3,3))
            for i in range(len(selectedIds)):
                inertia += masses[i]*np.outer(pos[i],pos[i])

            inertia /= np.sum(masses)

            # Find the largest inertia moment
            eigVal, eigVec = np.linalg.eig(inertia)
            maxInertia = np.argmax(eigVal)

            maxInertiaVal = eigVal[maxInertia]
            maxInertiaVec = eigVec[:,maxInertia]
            maxInertiaVec /= np.linalg.norm(maxInertiaVec)

            #Compare maxInertiaVec with vector, and if they are not the same, rotate
            if np.dot(maxInertiaVec,vector) < 0.9999:
                # Find the rotation axis
                rotAxis = np.cross(maxInertiaVec,vector)
                rotAxis /= np.linalg.norm(rotAxis)

                # Find the rotation angle
                rotAngle = np.arccos(np.dot(maxInertiaVec,vector))
                rot = R.from_rotvec(rotAngle*rotAxis)

                # Rotate the model, create rotation matrix using scipy
                center = np.sum(pos*masses[:,np.newaxis],axis=0)/np.sum(masses)
                pos -= center
                pos = rot.apply(pos)
                pos += center

                self.setIdsState(selectedIds,"position",pos.tolist())



import logging
import copy

import numpy as np

from scipy.spatial.transform import Rotation as R
from pyquaternion import Quaternion

from .bounds import *
from .particlesDistribution import *

#Geometry utils

def getEx(q):
    """ Given a quaternion, q, the function returns the z vector of the local basis"""

    q0,q1,q2,q3 = q

    return 2.0*np.asarray([q0*q0+q1*q1-0.5,q1*q2+q0*q3,q1*q3-q0*q2])

def getEy(q):
    """ Given a quaternion, q, the function returns the z vector of the local basis"""

    q0,q1,q2,q3 = q

    return 2.0*np.asarray([q1*q2-q0*q3,q0*q0+q2*q2-0.5,q2*q3+q0*q1])

def getEz(q):
    """ Given a quaternion, q, the function returns the z vector of the local basis"""

    q0,q1,q2,q3 = q

    return 2.0*np.asarray([q1*q3+q0*q2,q2*q3-q0*q1,q0*q0+q3*q3-0.5])


def quaternionFromVectors(vec1, vec2):
    """ Given two vector the function returns the rotation that transform one into another.
        The rotation is codified as a quaternion"""

    a, b = (vec1 / np.linalg.norm(vec1)).reshape(3), (vec2 / np.linalg.norm(vec2)).reshape(3)
    v = np.cross(a, b)

    if any(v):
        c = np.dot(a, b)
        s = np.linalg.norm(v)
        kmat = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
        M = np.eye(3) + kmat + kmat.dot(kmat) * ((1 - c) / (s ** 2))
    else:
        if np.dot(a,b) < 0:
            M = -np.eye(3)
        else:
            M =  np.eye(3)

    q1,q2,q3,q0 = R.from_matrix(M).as_quat()

    return np.asarray([q0,q1,q2,q3])




import sys, os
import copy
import json

import logging

import itertools

from . import modelBase
from ...utils.input import getSubParameters

import math

import numpy as np
from scipy.spatial.transform import Rotation

class MADna(modelBase):
    """
    Component name: MADna
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 17/03/2023

    MADna model for DNA. See https://pubs.acs.org/doi/pdf/10.1021/acs.jctc.2c00138

    :param sequence: DNA sequence
    :type sequence: str
    :param inputModelData: Path to the model data file
    :type inputModelData: str, optional. Default: "./data/MADna.json"
    :param debyeLength: Debye length
    :type debyeLength: float, optional. Default: 10.8
    :param dielectricConstant: Dielectric constant
    :type dielectricConstant: float, optional. Default: 78.3
    :param debyeFactor: Debye factor
    :type debyeFactor: float, optional. Default: 4.0
    :param variant: variant of the model. Options: "fast"
    :type variant: str, optional. Default: None
    """

    def __alignBasePairs(self,ref,mobile):

        refSel    = ref[6:10]
        mobileSel = mobile[0:4]

        refSelCentroid    = np.mean(refSel,axis=0)
        mobileSelCentroid = np.mean(mobileSel,axis=0)

        refSel    = refSel - refSelCentroid
        mobileSel = mobileSel - mobileSelCentroid

        ##### Kabsch #####
        C = np.dot(np.transpose(mobileSel),refSel)
        V, S, W = np.linalg.svd(C)
        if (np.linalg.det(V) * np.linalg.det(W)) < 0.0:
            S[-1] = -S[-1]
            V[:, -1] = -V[:, -1]
        U = np.dot(V, W)
        ##################

        mobile = mobile.__iadd__(-mobileSelCentroid)
        mobile = np.dot(mobile,U,out=mobile)
        mobile = mobile.__iadd__(refSelCentroid)

    def __update(self,iseq,loc):

        iS_5_1 = iseq*3
        iB_5_1 = iS_5_1 + 1

        iS_5_2 = self.nAtoms - 1 - iS_5_1 - 1
        iB_5_2 = self.nAtoms - 1 - iB_5_1 + 1

        iP_1 = iS_5_1 + 2
        iP_2 = self.nAtoms - 1 - iP_1

        iS_3_1 = iS_5_1 + 3
        iB_3_1 = iS_5_1 + 4

        iS_3_2 = self.nAtoms - 1 - iS_3_1 - 1
        iB_3_2 = self.nAtoms - 1 - iB_3_1 + 1

        #print(iseq,self.nAtoms)
        #print(iS_5_1,iB_5_1,iB_5_2,iS_5_2)
        #print(" ",iP_1,iP_2)
        #print(iS_3_1,iB_3_1,iB_3_2,iS_3_2)
        #input()

        #S B 5

        if iseq==0:
            self.coordinates[iS_5_1] = loc["S_5_1"]
            self.coordinates[iB_5_1] = loc["B_5_1"]
            self.coordinates[iB_5_2] = loc["B_5_2"]
            self.coordinates[iS_5_2] = loc["S_5_2"]

            self.structure[iS_5_1] = {"type":"S","strand":1,"basePairIndex":iseq+1}
            self.structure[iB_5_1] = {"type":self.seq[iseq],"strand":1,"basePairIndex":iseq+1}
            self.structure[iB_5_2] = {"type":self.model["PAIRS"][self.seq[iseq]],"strand":2,"basePairIndex":iseq+1}
            self.structure[iS_5_2] = {"type":"S","strand":2,"basePairIndex":iseq+1}
        else:
            self.coordinates[iS_5_1] = 0.5*(loc["S_5_1"] + self.coordinates[iS_5_1])
            self.coordinates[iB_5_1] = 0.5*(loc["B_5_1"] + self.coordinates[iB_5_1])
            self.coordinates[iB_5_2] = 0.5*(loc["B_5_2"] + self.coordinates[iB_5_2])
            self.coordinates[iS_5_2] = 0.5*(loc["S_5_2"] + self.coordinates[iS_5_2])


        self.coordinates[iS_3_1] = loc["S_3_1"]
        self.coordinates[iB_3_1] = loc["B_3_1"]
        self.coordinates[iB_3_2] = loc["B_3_2"]
        self.coordinates[iS_3_2] = loc["S_3_2"]

        self.coordinates[iP_1]   = loc["P_1"]
        self.coordinates[iP_2]   = loc["P_2"]

        #######################################################################

        self.structure[iP_1]   = {"type":"P","strand":1,"basePairIndex":iseq+2}
        self.structure[iP_2]   = {"type":"P","strand":2,"basePairIndex":iseq+1} #Note P 2 belongs to previous basePair

        self.structure[iS_3_1] = {"type":"S","strand":1,"basePairIndex":iseq+2}
        self.structure[iB_3_1] = {"type":self.seq[iseq+1],"strand":1,"basePairIndex":iseq+2}
        self.structure[iB_3_2] = {"type":self.model["PAIRS"][self.seq[iseq+1]],"strand":2,"basePairIndex":iseq+2}
        self.structure[iS_3_2] = {"type":"S","strand":2,"basePairIndex":iseq+2}

        return np.asarray([self.coordinates[iS_5_1], self.coordinates[iB_5_1], self.coordinates[iB_5_2], self.coordinates[iS_5_2],
                           self.coordinates[iP_1],   self.coordinates[iP_2],
                           self.coordinates[iS_3_1], self.coordinates[iB_3_1], self.coordinates[iB_3_2], self.coordinates[iS_3_2]])


    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"sequence",
                                                "inputModelData",
                                                "debyeLength","dielectricConstant",
                                                "debyeFactor",
                                                "variant"},
                         requiredParameters  = {"sequence"},
                         definedSelections   = {"type","strand","basePairIndex","basePairType","particleId"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        units = self.getUnits()

        if units.getName() != "KcalMol_A":
            self.logger.error(f"[MADna] Units are not set correctly. Please set units to \"KcalMol_A\" (selected: {units.getName()})")
            raise Exception("Not correct units")

        ############################################################

        self.modelData = params.get("inputModelData","./data/MADna.json")
        self.modelData = os.path.join(os.path.dirname(os.path.realpath(__file__)),self.modelData)

        with open(self.modelData,"r") as f:
            self.model = json.load(f)

        self.seq    = params.get("sequence")
        self.seqLen = len(self.seq)
        self.nAtoms = self.seqLen*6-2


        self.debyeLength        = params.get("debyeLength",10.8)
        self.dielectricConstant = params.get("dielectricConstant",78.3)

        self.debyeFactor = params.get("debyeFactor",4.0)

        self.variantName, self.variantParams = getSubParameters("variant",params)
        if   self.variantName is None:
            pass
        elif self.variantName == "fast":

            variantAvailableParameters = {"fastFactor"}
            variantRequiredParameters  = set()

            for key in self.variantParams.keys():
                if key not in variantAvailableParameters:
                    self.logger.error(f"[MADna] Variant parameter {key} is not available for variant {self.variantName}")
                    raise Exception("Not correct variant parameter")

            for key in variantRequiredParameters:
                if key not in self.variantParams.keys():
                    self.logger.error(f"[MADna] Variant parameter {key} is required for variant {self.variantName}")
                    raise Exception("Not correct variant parameter")

            self.fastFactor  = params.get("fastFactor",1)
        else:
            self.logger.error(f"[MADna] Variant {self.variantName} is not available")
            raise Exception("Variant not available")

        self.__generateCoordinatesAndTopology()

    def processSelection(self,**selection):

        indices = []
        for selType,selInfo in selection.items():
            #print(selType,selInfo,"type",type(selType),type(selInfo))
            if type(selInfo) != list:
                selection[selType]=[selInfo]

        idsSel      = set()
        typeSel     = set()
        strandSel   = set()
        basePairIndexSel = set()
        basePairTypeSel  = set()

        if "particleId" in selection.keys():
            for p in selection["particleId"]:
                idsSel.add(p)
        else:
            idsSel=set(range(self.nAtoms))

        if "type" in selection.keys():
            for t in selection["type"]:
                if t == "B":
                    typeSel.add("A")
                    typeSel.add("C")
                    typeSel.add("G")
                    typeSel.add("T")
                else:
                    typeSel.add(t)
        else:
            typeSel=set(["S","P","A","C","G","T"])

        if "strand" in selection.keys():
            for s in selection["strand"]:
                strandSel.add(s)
        else:
            strandSel=set([1,2])

        if "basePairIndex" in selection.keys():
            for bpi in selection["basePairIndex"]:
                if bpi > self.seqLen or bpi == 0:
                    self.logger.error(f"[MADna] Base pair has to be a number between 1 and {self.seqLen}. But is:{bpi}")
                    raise Exception("Base pair index out of range")
                if bpi < 0:
                    bpi = self.seqLen + bpi + 1
                basePairIndexSel.add(bpi)
        else:
            basePairIndexSel = set(list(range(self.seqLen)))

        if "basePairType" in selection.keys():
            for bpt in selection["basePairType"]:
                basePairTypeSel.add(bpt)
        else:
            basePairTypeSel = set(["A","C","G","T"])

        for info in self.structure.values():
            i=info["index"]
            t=info["type"]
            s=info["strand"]
            bpi=info["basePairIndex"]
            bpt=info["basePairType"]
            if (i in idsSel) and (t in typeSel) and (s in strandSel) and (bpi in basePairIndexSel) and (bpt in basePairTypeSel):
                indices.append(info["index"])

        return list(set(indices))

    def __generateCoordinatesAndTopology(self):

        self.coordinates = np.zeros((self.nAtoms,3))

        self.structure   = {}

        self.bonds       = []
        self.angles      = []
        self.dihedrals   = []

        self.exclusions  = []

        #######################################################
        ############ COORDINATES GENERATION STARTS ############

        nextCoord = np.zeros((10,3))

        loc = copy.deepcopy(self.model["POSITIONS"][self.seq[0:2]])

        r0 = np.asarray(loc["S_5_1"])
        for loc_el in loc:
            loc[loc_el] = np.asarray(loc[loc_el]) - r0

        prevCoord = self.__update(0,loc)

        for iseq in range(1,self.seqLen-1):
            loc = copy.deepcopy(self.model["POSITIONS"][self.seq[iseq:(iseq+2)]])

            for i,loc_el in enumerate(loc):
                nextCoord[i] = loc[loc_el]

            U = self.__alignBasePairs(prevCoord,nextCoord)

            loc = {'S_5_1': nextCoord[0], 'B_5_1': nextCoord[1], 'B_5_2': nextCoord[2], 'S_5_2': nextCoord[3],
                   'P_1':   nextCoord[4], 'P_2':   nextCoord[5],
                   'S_3_1': nextCoord[6], 'B_3_1': nextCoord[7], 'B_3_2': nextCoord[8], 'S_3_2': nextCoord[9]}

            prevCoord = self.__update(iseq,loc)

        for index,info in self.structure.items():
            info.update({"index":index})
            info.update({"basePairType":self.seq[self.structure[index]["basePairIndex"]-1]})

        #for index,info in self.structure.items():
        #    print(index,info)

        #Align chain along z axis

        p0 = self.coordinates[0]
        pf = self.coordinates[self.nAtoms//2-2]

        dr = pf - p0
        dr = dr/np.linalg.norm(dr)

        z  = np.array([0.0,0.0,1.0])

        axis = np.cross(dr,z)
        axis = axis/np.linalg.norm(axis)

        angle = np.arccos(np.dot(dr,z))

        rot = Rotation.from_rotvec(angle * axis)
        self.coordinates = rot.apply(self.coordinates)

        #Make overall translation such that sugar center of second-to-last bp has x=0, y=0

        p0 = self.coordinates[self.nAtoms//2-5]
        pf = self.coordinates[self.nAtoms//2+3]

        trans = 0.5*(p0 + pf)
        for i in range(0,self.nAtoms):
            self.coordinates[i] -= trans


        ############# COORDINATES GENERATION ENDS #############
        #######################################################

        #######################################################
        ############# TOPOLOGY GENERATION STARTS ##############

        basePairsList = list(range(1,self.seqLen+1))

        basePairsPairs = []
        for i,bp in enumerate(basePairsList[:-1]):
            basePairsPairs.append([basePairsList[i],basePairsList[i+1]])

        #Same base pair
        for bp in basePairsList:
            t = self.seq[bp-1]
            ##SB
            i,j = self.processSelection(**{"basePairIndex":bp,"type":["S",t],"strand":1})
            bType = "SB/{}".format(t)
            self.bonds.append({"i":i,"j":j,"type":bType})

            i,j = self.processSelection(**{"basePairIndex":bp,"type":["S",self.model["PAIRS"][t]],"strand":2})
            bType = "SB/{}".format(self.model["PAIRS"][t])
            self.bonds.append({"i":j,"j":i,"type":bType})#Note index inverted!!

            ##BB-inter strand
            i,j = self.processSelection(**{"basePairIndex":bp,"type":[t,self.model["PAIRS"][t]]})
            bType = "BB-inter/{}{}".format(t,self.model["PAIRS"][t])
            self.bonds.append({"i":i,"j":j,"type":bType})

        for bpp in basePairsPairs:
            bp5,bp3 = bpp
            t5,t3 = self.seq[bp5-1],self.seq[bp3-1]

            i = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            if len(i)==1 and len(j)==1:
                i=i[0]
                j=j[0]

                #print(bp5,bp3,i,j,1)

                bType = "SP/{}{}".format(t5,t3)
                self.bonds.append({"i":i,"j":j,"type":bType})

            i = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":2})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            if len(i)==1 and len(j)==1:
                i=i[0]
                j=j[0]

                #print(bp5,bp3,i,j,1)

                bType = "PS/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.bonds.append({"i":j,"j":i,"type":bType}) #Note index inverted!!

            #########################################################################

            i = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})
            if len(i)==1 and len(j)==1:
                i=i[0]
                j=j[0]

                #print(bp5,bp3,t5,t3,i,j)

                bType = "PS/{}{}".format(t5,t3)
                self.bonds.append({"i":i,"j":j,"type":bType})

            i = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":2})
            if len(i)==1 and len(j)==1:
                i=i[0]
                j=j[0]

                #print(bp5,bp3,i,j,2)

                bType = "SP/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.bonds.append({"i":j,"j":i,"type":bType}) #Note index inverted!!

            ########################################################################

            #5BB3-intra strand
            i = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":1})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":1})
            if len(i)==1 and len(j)==1:
                i=i[0]
                j=j[0]

                bType = "BB-intra/{}{}".format(t5,t3)
                self.bonds.append({"i":i,"j":j,"type":bType})

            #5BB3-intra strand
            i = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":2})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":2})
            if len(i)==1 and len(j)==1:
                i=i[0]
                j=j[0]

                bType = "BB-intra/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.bonds.append({"i":j,"j":i,"type":bType}) #Note index inverted!!

        #########################################################################

        #Same base pair
        for bp in basePairsList:
            t = self.seq[bp-1]
            ##SBB
            i, = self.processSelection(**{"basePairIndex":bp,"type":"S","strand":1})
            j, = self.processSelection(**{"basePairIndex":bp,"type":"B","strand":1})
            k, = self.processSelection(**{"basePairIndex":bp,"type":"B","strand":2})

            aType = "SBB/{}{}".format(t,self.model["PAIRS"][t])
            self.angles.append({"i":i,"j":j,"k":k,"type":aType})

            i, = self.processSelection(**{"basePairIndex":bp,"type":"B","strand":1})
            j, = self.processSelection(**{"basePairIndex":bp,"type":"B","strand":2})
            k, = self.processSelection(**{"basePairIndex":bp,"type":"S","strand":2})

            aType = "SBB/{}{}".format(self.model["PAIRS"][t],t)
            self.angles.append({"i":k,"j":j,"k":i,"type":aType})#Note index inverted!!

        #########################################################################

        #Diferent base pair
        for bpp in basePairsPairs:
            bp5,bp3 = bpp
            t5,t3 = self.seq[bp5-1],self.seq[bp3-1]

            ########################################################################

            i = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                aType = "SPS/{}{}".format(t5,t3)
                self.angles.append({"i":i,"j":j,"k":k,"type":aType})

            i = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":2})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1:
                i = i[0]
                j = j[0]
                k = k[0]

                aType = "SPS/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.angles.append({"i":k,"j":j,"k":i,"type":aType})

            ########################################################################

            i = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
            k = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                aType = "3PSB5/{}{}".format(t5,t3)
                self.angles.append({"i":i,"j":j,"k":k,"type":aType})

            i = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":2})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                aType = "3PSB5/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.angles.append({"i":i,"j":j,"k":k,"type":aType})

            ########################################################################

            i = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                aType = "5PSB3/{}{}".format(t5,t3)
                self.angles.append({"i":i,"j":j,"k":k,"type":aType})

            i = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":2})
            k = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                aType = "5PSB3/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.angles.append({"i":i,"j":j,"k":k,"type":aType})

        #########################################################################

        for bpp in basePairsPairs:
            bp5,bp3 = bpp
            t5,t3 = self.seq[bp5-1],self.seq[bp3-1]

            #PSBB53
            i = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":1})
            l = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "PSBB53/{}{}".format(t5,t3)
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ##PSBB53
            i = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":2})
            k = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":2})
            l = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "PSBB53/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ########################################################################

            ###PSBB35
            i = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
            k = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":1})
            l = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "PSBB35/{}{}".format(t5,t3)
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ###PSBB35
            i = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":2})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":2})
            l = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "PSBB35/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ########################################################################

            ##SPSP
            if bp3+1 <= self.seqLen:
                i = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
                j = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
                k = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})
                l = self.processSelection(**{"basePairIndex":bp3+1,"type":"P","strand":1})

                if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                    i = i[0]
                    j = j[0]
                    k = k[0]
                    l = l[0]

                    dType = "SPSP/{}{}".format(t5,t3)
                    self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ##SPSP
            if bp5-1 >= 1:
                i = self.processSelection(**{"basePairIndex":bp3  ,"type":"S","strand":2})
                j = self.processSelection(**{"basePairIndex":bp5  ,"type":"P","strand":2})
                k = self.processSelection(**{"basePairIndex":bp5  ,"type":"S","strand":2})
                l = self.processSelection(**{"basePairIndex":bp5-1,"type":"P","strand":2})

                if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                    i = i[0]
                    j = j[0]
                    k = k[0]
                    l = l[0]

                    dType = "SPSP/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                    self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ########################################################################

            ##PSPS
            i = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":1})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            l = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "PSPS/{}{}".format(t5,t3)
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ##PSPS
            i = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":2})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":2})
            k = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            l = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "PSPS/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ########################################################################

            ##SPSB53
            i = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})
            l = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "SPSB53/{}{}".format(t5,t3)
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ##SPSB53
            i = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":2})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            k = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":2})
            l = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "SPSB53/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ########################################################################

            ##SPSB35
            i = self.processSelection(**{"basePairIndex":bp5,"type":"B","strand":1})
            j = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":1})
            k = self.processSelection(**{"basePairIndex":bp3,"type":"P","strand":1})
            l = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":1})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "SPSB35/{}{}".format(t5,t3)
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ##SPSB35
            i = self.processSelection(**{"basePairIndex":bp3,"type":"B","strand":2})
            j = self.processSelection(**{"basePairIndex":bp3,"type":"S","strand":2})
            k = self.processSelection(**{"basePairIndex":bp5,"type":"P","strand":2})
            l = self.processSelection(**{"basePairIndex":bp5,"type":"S","strand":2})

            if len(i)==1 and len(j)==1 and len(k)==1 and len(l)==1:
                i = i[0]
                j = j[0]
                k = k[0]
                l = l[0]

                dType = "SPSB35/{}{}".format(self.model["PAIRS"][t3],self.model["PAIRS"][t5])
                self.dihedrals.append({"i":i,"j":j,"k":k,"l":l,"type":dType})

            ########################################################################

        for i in range(self.nAtoms):
            self.exclusions.append([])

        neighbours = []
        for i in range(self.nAtoms):
            neighbours.append([])

        for bnd in self.bonds:
            neighbours[bnd["i"]].append(bnd["j"])
            neighbours[bnd["j"]].append(bnd["i"])

        for i in range(self.nAtoms):
            currentNeig = [i]
            for dst in range(1,4):
                nextNeig = []
                for i0 in currentNeig:
                    for j in neighbours[i0]:
                        nextNeig.append(j)
                        self.exclusions[i].append(j)
                currentNeig = nextNeig

        for i in range(self.nAtoms):
            self.exclusions[i] = list(set(self.exclusions[i]))
            self.exclusions[i].sort(key=int)
            if i in self.exclusions[i]:
                self.exclusions[i].remove(i)

        ############### TOPOLOGY GENERATION ENDS ##############
        #######################################################

        #Add particle types
        types = self.getTypes()

        for name,t in self.model["TYPES"].items():
            types.addType(name=name,**t)

        state  = {}
        state["labels"] = ["id","position"]
        state["data"]   = []
        for i,pos in enumerate(self.coordinates):
            state["data"].append([i,[pos[0],pos[1],pos[2]]])

        struct = {}
        struct["labels"] = ["id","type","resId","chainId"]
        struct["data"] = []
        for index,info in self.structure.items():

            typ = info["type"]
            res = info["basePairIndex"]
            ch  = info["strand"]

            struct["data"].append([index,typ,res,ch])

        forceField = {}

        #NL
        forceField["nl"] = {}
        forceField["nl"]["type"]       = ["VerletConditionalListSet", "nonExcluded"]
        forceField["nl"]["parameters"] = {"cutOffVerletFactor": 1.5}
        forceField["nl"]["labels"]     = ["id", "id_list"]
        forceField["nl"]["data"] = []

        for i in range(self.nAtoms):
            forceField["nl"]["data"].append([i,self.exclusions[i]])

        #WCA
        forceField["WCA"] = {}
        forceField["WCA"]["type"]       = ["NonBonded", "WCAType2"]
        forceField["WCA"]["parameters"] = {"cutOffFactor": 2.5,
                                           "condition":"nonExcluded"}

        forceField["WCA"]["labels"] = ["name_i", "name_j", "epsilon", "sigma"]
        forceField["WCA"]["data"]   = []

        for t1,info1 in self.model["TYPES"].items():
            for t2,info2 in self.model["TYPES"].items():
                forceField["WCA"]["data"].append([t1,t2,1.0,info1["radius"]+info2["radius"]])

        #DH
        forceField["DH"] = {}
        forceField["DH"]["type"]       = ["NonBonded", "DebyeHuckel"]
        forceField["DH"]["parameters"] = {"cutOffFactor": self.debyeFactor,
                                          "debyeLength":self.debyeLength,
                                          "dielectricConstant":self.dielectricConstant,
                                          "condition":"nonExcluded"}

        #BONDS
        forceField["BONDS"] = {}
        forceField["BONDS"]["type"]   = ["Bond2", "Harmonic"]
        forceField["BONDS"]["parameters"] = {}
        forceField["BONDS"]["labels"]     = ["id_i", "id_j", "K", "r0"]
        forceField["BONDS"]["data"]       = []

        for bnd in self.bonds:
            forceField["BONDS"]["data"].append([bnd["i"],bnd["j"],
                                                self.model["BONDS"][bnd["type"]]["K"],
                                                self.model["BONDS"][bnd["type"]]["r0"]])

        #ANGLES
        forceField["ANGLES"] = {}
        forceField["ANGLES"]["type"]   = ["Bond3", "HarmonicAngular"]
        forceField["ANGLES"]["parameters"] = {}
        forceField["ANGLES"]["labels"]     = ["id_i", "id_j", "id_k", "K", "theta0"]
        forceField["ANGLES"]["data"]       = []

        for ang in self.angles:
            forceField["ANGLES"]["data"].append([ang["i"],ang["j"],ang["k"],
                                                self.model["ANGLES"][ang["type"]]["K"],
                                                self.model["ANGLES"][ang["type"]]["theta0"]])

        #DIHEDRALS
        forceField["DIHEDRALS"] = {}
        forceField["DIHEDRALS"]["type"]   = ["Bond4", "Dihedral"]
        forceField["DIHEDRALS"]["parameters"] = {}
        forceField["DIHEDRALS"]["labels"]     = ["id_i", "id_j", "id_k", "id_l", "n", "K", "phi0"]
        forceField["DIHEDRALS"]["data"]       = []

        for dih in self.dihedrals:
            forceField["DIHEDRALS"]["data"].append([dih["i"],dih["j"],dih["k"],dih["l"],
                                                    1,
                                                    self.model["DIHEDRALS"][dih["type"]]["K"],
                                                    self.model["DIHEDRALS"][dih["type"]]["phi0"]])

        #######################################################
        ####################### VARIANT #######################

        if self.variantName == "fast":

            del forceField["WCA"]
            del forceField["DH"]

            PHOSPHATE_DISTANCE = 3.4

            cutOff      = self.debyeFactor*self.debyeLength
            madnafast_n = math.ceil(self.fastFactor*(cutOff/PHOSPHATE_DISTANCE))

            phosphateIndex_basePair = []
            for bp in range(self.seqLen):
                index = self.processSelection(**{"basePairIndex":bp+1,"type":"P"})
                phosphateIndex_basePair+=[[i,bp+1] for i in index]

            forceField["BONDS_DH"] = {}
            forceField["BONDS_DH"]["type"]       = ["Bond2", "DebyeHuckel"]
            forceField["BONDS_DH"]["parameters"] = {}
            forceField["BONDS_DH"]["labels"]     = ["id_i", "id_j", "chgProduct", "dielectricConstant", "debyeLength", "cutOff"]
            forceField["BONDS_DH"]["data"]       = []

            chgProduct = self.model["TYPES"]["P"]["charge"]
            chgProduct = chgProduct*chgProduct
            for ph1,ph2 in itertools.combinations(phosphateIndex_basePair,2):
                if(abs(ph1[1]-ph2[1])<madnafast_n):
                    forceField["BONDS_DH"]["data"].append([ph1[0],ph2[0],chgProduct,self.dielectricConstant,self.debyeLength,cutOff])

        ##################### VARIANT END #####################
        #######################################################

        self.setState(state)
        self.setStructure(struct)
        self.setForceField(forceField)




from VLMP.components.models import modelBase

class PARTICLE(modelBase):
    """
    Component name: PARTILCE
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 18/06/2023

    Single particle model.

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name= name,
                         availableParameters = {"particleName",
                                                "particleMass","particleRadius","particleCharge",
                                                "position"},
                         requiredParameters  = {"particleName"},
                         definedSelections   = {"particleId"},
                         **params)

        ############################################################

        particleName = params["particleName"]

        particleMass   = params.get("particleMass",1.0)
        particleRadius = params.get("particleRadius",1.0)
        particleCharge = params.get("particleCharge",0.0)

        types = self.getTypes()
        types.addType(name = particleName,
                      mass = particleMass,
                      radius = particleRadius,
                      charge = particleCharge)

        state = {}
        state["labels"] = ["id","position"]

        particlePosition = params.get("position",[0.0,0.0,0.0])
        state["data"] = [[0,particlePosition]]

        structure = {}
        structure["labels"] = ["id","type"]
        structure["data"] = [[0,particleName]]

        ############################################################

        self.setState(state)
        self.setStructure(structure)


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += [0]

        return sel




import os
import copy

import logging

################ UNITS INTERFACE ################

from pyUAMMD import simulation

class unitsBase:

    def __init__(self,
                 _type:str,_name:str,
                 availableParameters:set,
                 requiredParameters:set,
                 **params):

        self.logger = logging.getLogger("VLMP")

        self._type = _type
        self._name = _name

        self.availableParameters = availableParameters.copy()
        self.requiredParameters  = requiredParameters.copy()

        # Check all required parameters are available parameters
        if not self.requiredParameters.issubset(self.availableParameters):
            notAvailable = self.requiredParameters.difference(self.availableParameters)
            self.logger.error(f"[Units] ({self._type}) Some required parameters ({notAvailable}) are not available parameters for units {self._name}")
            raise Exception(f"Required paramaters are not available parameters")

        # Check if all parameters given by params are available
        for par in params:
            if par not in self.availableParameters:
                self.logger.error(f"[Units] ({self._type}) Parameter {par} not available for units {self._name}")
                raise Exception(f"Parameter not available")

        # Check if all required parameters are given
        for par in self.requiredParameters:
            if par not in params:
                self.logger.error(f"[Units] ({self._type}) Required parameter {par} not given for units {self._name}")
                raise Exception(f"Required parameter not given")

        self.logger.info(f"[Units] ({self._type}) Using units {self._name}")

        ########################################################

        self._unitsUAMMD = None
        self._constants  = None

    ########################################################

    def getName(self):
        return self._name

    def getType(self):
        return self._type

    ########################################################

    def setUnitsName(self, unitsUAMMD):
        self._unitsUAMMD = unitsUAMMD

    def getUnitsName(self):
        if self._unitsUAMMD is None:
            self.logger.error(f"[Units] ({self._type}) Units not set for units {self._name}")
            raise Exception(f"Units not set")
        return self._unitsUAMMD

    def addConstant(self, constName, constValue):
        if self._constants is None:
            self._constants = {}
        if constName in self._constants.keys():
            self.logger.error(f"[Units] ({self._type}) Constant {constName} already add for units {self._name}")
            raise Exception(f"Constant already added")
        self._constants[constName] = constValue

    def getConstant(self, constName):
        if self._constants is None:
            self.logger.error(f"[Units] ({self._type}) No constants added for units {self._name}")
            raise Exception(f"No constants added")
        if constName not in self._constants.keys():
            self.logger.error(f"[Units] ({self._type}) Constant {constName} not added for units {self._name}")
            raise Exception(f"Constant not added")
        return self._constants[constName]

    ########################################################

    def getSimulation(self,DEBUG_MODE = False):
        return simulation({"global":{"units":{"type":["Units",self.getUnitsName()]}}},DEBUG_MODE)

############### IMPORT ALL UNITS ###############

import glob

currentPath = os.path.dirname(os.path.abspath(__file__))
units = [os.path.basename(x) for x in glob.glob(currentPath+"/*.py")]
units = [x.split(".")[0] for x in units if "__" not in x]

for u in units:
    try:
        exec(f"from .{u} import *")
    except Exception as e:
        logging.getLogger("VLMP").error(e)
        logging.getLogger("VLMP").error(f"[Units] Error importing units type component {u}")



import sys, os

import logging

from . import modelBase

import numpy as np

class WLC(modelBase):
    """
    Component name: WLC
    Component type: model

    Author: Pablo Ibáñez-Freire
    Date: 13/03/2023

    Worm-like chain model. See https://en.wikipedia.org/wiki/Worm-like_chain

    :param N: Number of particles
    :type N: int
    :param mass: Mass of the particles
    :type mass: float, optional. Default: 1.0
    :param b: Distance between two consecutive particles
    :type b: float, optional. Default: 1.0
    :param Kb: Spring constant for bonds
    :type Kb: float, optional. Default: 1.0
    :param Ka: Spring constant for angles
    :type Ka: float, optional. Default: 1.0
    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"N","mass","b","Kb","Ka","typeName"},
                         requiredParameters  = {"N"},
                         definedSelections   = {"particleId","polymerIndex"},
                         **params)

        ############################################################
        ####################  Model Parameters  ####################
        ############################################################

        #Read the number of particles
        self.N = params["N"]

        #Read the particle mass and the distance between two consecutive particles
        mass = params.get("mass",1.0)
        b    = params.get("b",1.0)

        #Read values for the spring constant for bonds and angles
        Kb = params.get("Kb",1.0)
        Ka = params.get("Ka",1.0)

        typeName = params.get("typeName","A")

        self.logger.debug(f"[WLC] Generating a WLC model with {self.N} particles")
        self.logger.debug(f"[WLC] Parameters: mass={mass}, b={b}, Kb={Kb}, Ka={Ka}")

        ########################################################
        ############### Generate the WLC model #################
        ########################################################

        #Add particle types
        types = self.getTypes()
        radius = 0.5*b
        types.addType(name=typeName,mass=mass,radius=radius)

        #Generate positions, a line along the z axis
        state = {}
        state["labels"] = ["id","position"]
        state["data"]   = []
        for i in range(self.N):
            state["data"].append([i, [0.0,0.0,i*b-0.5*(self.N-1)*b]])

        #Generate structure
        structure = {}
        structure["labels"] = ["id","type"]
        structure["data"]   = []
        for i in range(self.N):
            structure["data"].append([i,typeName])

        #Generate forceField
        forceField = {}

        forceField["bonds_wlc"] = {}
        forceField["bonds_wlc"]["type"]       = ["Bond2","Harmonic"]
        forceField["bonds_wlc"]["parameters"] = {}
        forceField["bonds_wlc"]["labels"]     = ["id_i","id_j","K","r0"]
        forceField["bonds_wlc"]["data"]       = []

        for i in range(self.N-1):
            forceField["bonds_wlc"]["data"].append([i,i+1,Kb,b])

        forceField["angles_wlc"] = {}
        forceField["angles_wlc"]["type"]       = ["Bond3","KratkyPorod"]
        forceField["angles_wlc"]["parameters"] = {}
        forceField["angles_wlc"]["labels"]     = ["id_i","id_j","id_k","K"]
        forceField["angles_wlc"]["data"]       = []

        for i in range(self.N-2):
            forceField["angles_wlc"]["data"].append([i,i+1,i+2,Ka])

        ########################################################

        self.setState(state)
        self.setStructure(structure)
        self.setForceField(forceField)


    def processSelection(self,**params):

        sel = []

        if "particleId" in params:
            sel += params["particleId"]

        if "polymerIndex" in params:
            for pIndex in params["polymerIndex"]:
                if pIndex < -self.N or pIndex > self.N or pIndex == 0:
                    self.logger.error(f"[WLC] Invalid polymer index {pIndex}")
                    raise Exception("Invalid polymer index")

                if pIndex > 0:
                    sel.append(pIndex-1)
                if pIndex < 0:
                    sel.append(self.N+pIndex)

        return sel



import sys, os

import logging

from . import simulationStepBase

class AFMMaxForce(simulationStepBase):
    """
    Component name: AFMMaxForce
    Component type: simulationStep

    Author: Pablo Ibáñez-Freire
    Date: 31/08/2023

    """

    def __init__(self,name,**params):
        super().__init__(_type= self.__class__.__name__,
                         _name= name,
                         availableParameters = {"maxForce"},
                         requiredParameters  = {"maxForce"},
                         availableSelections = set(),
                         requiredSelections  = set(),
                         **params)

        ############################################################
        ############################################################
        ############################################################

        maxForce = params["maxForce"]

        simulationStep = {
            name:{
                  "type":["FlowControl","AFMMaxForce"],
                  "parameters":{"maxForce":maxForce}
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)





import sys, os

import logging

from . import simulationStepBase

class nativeContactsMeasurement(simulationStepBase):
    """
    Component name: nativeContactsMeasurement
    Component type: simulationStep

    """

    def __init__(self,name,**params):
        super().__init__(_type = self.__class__.__name__,
                         _name = name,
                         availableParameters = {"outputFilePath"},
                         requiredParameters  = {"outputFilePath"},
                         availableSelections = {"selection"},
                         requiredSelections  = {"selection"},
                         **params)

        ############################################################
        ############################################################
        ############################################################

        parameters = {}

        parameters["outputFilePath"] = params["outputFilePath"]

        selIds = self.getSelection("selection")
        data = []
        for i,j in selIds:
            data.append([i,j])

        simulationStep = {
            name:{
              "type":["ParticlesListMeasure","ContactsMeasure"],
              "parameters":{**parameters},
              "labels":["id_i","id_j"],
              "data":data.copy()
            }
        }

        ############################################################

        self.setSimulationStep(simulationStep)



